from collections.abc import Sequence
from typing import NamedTuple, Optional, Union

from typing_extensions import Self, TypeAlias

import dagster._check as check
from dagster._core.definitions.asset_selection import (
    AssetSelection,
    CoercibleToAssetSelection,
    is_coercible_to_asset_selection,
)
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.unresolved_asset_job_definition import (
    UnresolvedAssetJobDefinition,
    define_asset_job,
)
from dagster._utils.warnings import deprecation_warning

ExecutableDefinition: TypeAlias = Union[
    JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition
]

CoercibleToAutomationTarget: TypeAlias = Union[
    CoercibleToAssetSelection,
    AssetsDefinition,
    ExecutableDefinition,
]

ResolvableToJob: TypeAlias = Union[JobDefinition, UnresolvedAssetJobDefinition, str]
"""
A piece of data that is resolvable to a JobDefinition. One of:

- JobDefinition
- UnresolvedAssetJobDefinition
- str (a job name)

The property of being resolvable to a JobDefinition is what unites all of the entities that an
AutomationTarget can wrap.
"""

ANONYMOUS_ASSET_JOB_PREFIX = "__anonymous_asset_job"


class AutomationTarget(
    NamedTuple(
        "_AutomationTarget",
        [
            ("resolvable_to_job", ResolvableToJob),
            ("op_selection", Optional[Sequence[str]]),
            ("assets_defs", Sequence[Union[AssetsDefinition, SourceAsset]]),
        ],
    )
):
    """An abstraction representing a job to be executed by an automation, i.e. schedule or sensor.

    Args:
        resolvable_to_job (ResolvableToJob): An entity that is resolvable to a job at
            definition-resolution time.
        op_selection (Optional[Sequence[str]]): An optional list of op names to execute within the job.
    """

    def __new__(
        cls,
        resolvable_to_job: Union[JobDefinition, UnresolvedAssetJobDefinition, str],
        op_selection: Optional[Sequence[str]] = None,
        assets_defs: Optional[Sequence[Union[AssetsDefinition, SourceAsset]]] = None,
    ):
        return super().__new__(
            cls,
            resolvable_to_job=check.inst_param(
                resolvable_to_job,
                "resolvable_to_job",
                (JobDefinition, UnresolvedAssetJobDefinition, str),
            ),
            op_selection=check.opt_nullable_sequence_param(
                op_selection, "op_selection", of_type=str
            ),
            assets_defs=check.opt_sequence_param(
                assets_defs, "assets_defs", of_type=AssetsDefinition
            ),
        )

    @classmethod
    def from_coercible(
        cls,
        coercible: CoercibleToAutomationTarget,
        automation_name: Optional[str] = None,  # only needed if we are generating an anonymous job
    ) -> Self:
        if isinstance(coercible, (JobDefinition, UnresolvedAssetJobDefinition)):
            return cls(coercible)
        elif isinstance(coercible, GraphDefinition):
            deprecation_warning(
                "Passing GraphDefinition as a job/target argument to ScheduleDefinition and SensorDefinition",
                breaking_version="2.0",
            )
            return cls(coercible.to_job())
        else:
            if isinstance(coercible, AssetsDefinition):
                asset_selection = AssetSelection.assets(coercible)
                assets_defs = [coercible]
            elif isinstance(coercible, AssetSelection):
                asset_selection = coercible
                assets_defs = []
            elif is_coercible_to_asset_selection(coercible):
                asset_selection = AssetSelection.from_coercible(coercible)
                assets_defs = (
                    [x for x in coercible if isinstance(x, (AssetsDefinition, SourceAsset))]
                    if isinstance(coercible, Sequence)
                    else []
                )
            else:
                check.failed(_make_invalid_target_error_message(coercible))

            job_name = _make_anonymous_asset_job_name(
                check.not_none(automation_name, "Must provide automation_name")
            )
            return cls(
                resolvable_to_job=define_asset_job(name=job_name, selection=asset_selection),
                assets_defs=assets_defs,
            )

    @property
    def job_name(self) -> str:
        if isinstance(self.resolvable_to_job, str):
            return self.resolvable_to_job
        else:
            return self.resolvable_to_job.name

    @property
    def job_def(self) -> Union[JobDefinition, UnresolvedAssetJobDefinition]:
        if isinstance(self.resolvable_to_job, str):
            check.failed(
                "Cannot access job_def for a target with string job name for resolvable_to_job"
            )
        return self.resolvable_to_job

    @property
    def has_job_def(self) -> bool:
        return isinstance(self.resolvable_to_job, (JobDefinition, UnresolvedAssetJobDefinition))


def _make_anonymous_asset_job_name(automation_name: str) -> str:
    return f"{ANONYMOUS_ASSET_JOB_PREFIX}_{automation_name}"


def _make_invalid_target_error_message(target: object) -> str:
    if isinstance(target, Sequence):
        additional_message = " If you pass a sequence to a schedule, it must be a sequence of strings, AssetKeys, AssetsDefinitions, or SourceAssets."
    else:
        additional_message = ""
    return f"Invalid target passed to schedule/sensor.{additional_message} Target: {target}"



import logging
import uuid
import zlib
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from enum import Enum
from typing import Any, Callable, ContextManager, NamedTuple, Optional, Union, cast  # noqa: UP035

import sqlalchemy as db
import sqlalchemy.exc as db_exc
from sqlalchemy.engine import Connection

import dagster._check as check
from dagster._core.errors import (
    DagsterInvariantViolationError,
    DagsterRunAlreadyExists,
    DagsterRunNotFoundError,
    DagsterSnapshotDoesNotExist,
)
from dagster._core.events import (
    EVENT_TYPE_TO_PIPELINE_RUN_STATUS,
    DagsterEvent,
    DagsterEventType,
    RunFailureReason,
)
from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus, PartitionBackfill
from dagster._core.remote_representation.origin import RemoteJobOrigin
from dagster._core.snap import ExecutionPlanSnapshot, JobSnap, create_execution_plan_snapshot_id
from dagster._core.storage.dagster_run import (
    DagsterRun,
    DagsterRunStatus,
    JobBucket,
    RunPartitionData,
    RunRecord,
    RunsFilter,
    TagBucket,
)
from dagster._core.storage.runs.base import RunStorage
from dagster._core.storage.runs.migration import (
    BACKFILL_JOB_NAME_AND_TAGS,
    OPTIONAL_DATA_MIGRATIONS,
    REQUIRED_DATA_MIGRATIONS,
    RUN_BACKFILL_ID,
    RUN_PARTITIONS,
    MigrationFn,
)
from dagster._core.storage.runs.schema import (
    BackfillTagsTable,
    BulkActionsTable,
    DaemonHeartbeatsTable,
    InstanceInfo,
    KeyValueStoreTable,
    RunsTable,
    RunTagsTable,
    SecondaryIndexMigrationTable,
    SnapshotsTable,
)
from dagster._core.storage.sql import SqlAlchemyQuery
from dagster._core.storage.sqlalchemy_compat import (
    db_fetch_mappings,
    db_scalar_subquery,
    db_select,
    db_subquery,
)
from dagster._core.storage.tags import (
    BACKFILL_ID_TAG,
    PARTITION_NAME_TAG,
    PARTITION_SET_TAG,
    REPOSITORY_LABEL_TAG,
    ROOT_RUN_ID_TAG,
    RUN_FAILURE_REASON_TAG,
)
from dagster._daemon.types import DaemonHeartbeat
from dagster._serdes import deserialize_value, serialize_value
from dagster._serdes.serdes import deserialize_values
from dagster._seven import JSONDecodeError
from dagster._time import datetime_from_timestamp, get_current_datetime, utc_datetime_from_naive
from dagster._utils import PrintFn
from dagster._utils.merger import merge_dicts


class SnapshotType(Enum):
    PIPELINE = "PIPELINE"
    EXECUTION_PLAN = "EXECUTION_PLAN"


class SqlRunStorage(RunStorage):
    """Base class for SQL based run storages."""

    @abstractmethod
    def connect(self) -> ContextManager[Connection]:
        """Context manager yielding a sqlalchemy.engine.Connection."""

    @abstractmethod
    def upgrade(self) -> None:
        """This method should perform any schema or data migrations necessary to bring an
        out-of-date instance of the storage up to date.
        """

    def fetchall(self, query: SqlAlchemyQuery) -> Sequence[Any]:
        with self.connect() as conn:
            return db_fetch_mappings(conn, query)

    def fetchone(self, query: SqlAlchemyQuery) -> Optional[Any]:
        with self.connect() as conn:
            if db.__version__.startswith("2."):
                return conn.execute(query).mappings().first()
            else:
                return conn.execute(query).fetchone()

    def add_run(self, dagster_run: DagsterRun) -> DagsterRun:
        check.inst_param(dagster_run, "dagster_run", DagsterRun)

        if dagster_run.job_snapshot_id and not self.has_job_snapshot(dagster_run.job_snapshot_id):
            raise DagsterSnapshotDoesNotExist(
                f"Snapshot {dagster_run.job_snapshot_id} does not exist in run storage"
            )

        has_tags = dagster_run.tags and len(dagster_run.tags) > 0
        partition = dagster_run.tags.get(PARTITION_NAME_TAG) if has_tags else None
        partition_set = dagster_run.tags.get(PARTITION_SET_TAG) if has_tags else None
        values = {
            "run_id": dagster_run.run_id,
            "pipeline_name": dagster_run.job_name,
            "status": dagster_run.status.value,
            "run_body": serialize_value(dagster_run),
            "snapshot_id": dagster_run.job_snapshot_id,
            "partition": partition,
            "partition_set": partition_set,
        }
        if self.has_backfill_id_column():
            values["backfill_id"] = dagster_run.tags.get(BACKFILL_ID_TAG)

        runs_insert = RunsTable.insert().values(**values)
        with self.connect() as conn:
            try:
                conn.execute(runs_insert)
            except db_exc.IntegrityError as exc:
                raise DagsterRunAlreadyExists from exc

            tags_to_insert = dagster_run.tags_for_storage()
            if tags_to_insert:
                conn.execute(
                    RunTagsTable.insert(),
                    [
                        dict(run_id=dagster_run.run_id, key=k, value=v)
                        for k, v in tags_to_insert.items()
                    ],
                )

        return dagster_run

    def handle_run_event(self, run_id: str, event: DagsterEvent) -> None:
        from dagster._core.events import JobFailureData

        check.str_param(run_id, "run_id")
        check.inst_param(event, "event", DagsterEvent)

        if event.event_type not in EVENT_TYPE_TO_PIPELINE_RUN_STATUS:
            return

        run = self._get_run_by_id(run_id)
        if not run:
            # TODO log?
            return

        new_job_status = EVENT_TYPE_TO_PIPELINE_RUN_STATUS[event.event_type]

        run_stats_cols_in_index = self.has_run_stats_index_cols()

        kwargs = {}

        # consider changing the `handle_run_event` signature to get timestamp off of the
        # EventLogEntry instead of the DagsterEvent, for consistency
        now = get_current_datetime()

        if run_stats_cols_in_index and event.event_type == DagsterEventType.PIPELINE_START:
            kwargs["start_time"] = now.timestamp()

        if run_stats_cols_in_index and event.event_type in {
            DagsterEventType.PIPELINE_CANCELED,
            DagsterEventType.PIPELINE_FAILURE,
            DagsterEventType.PIPELINE_SUCCESS,
        }:
            kwargs["end_time"] = now.timestamp()

        with self.connect() as conn:
            conn.execute(
                RunsTable.update()
                .where(RunsTable.c.run_id == run_id)
                .values(
                    run_body=serialize_value(run.with_status(new_job_status)),
                    status=new_job_status.value,
                    update_timestamp=now,
                    **kwargs,
                )
            )

        if event.event_type == DagsterEventType.PIPELINE_FAILURE and isinstance(
            event.event_specific_data, JobFailureData
        ):
            failure_reason = event.event_specific_data.failure_reason

            if failure_reason and failure_reason != RunFailureReason.UNKNOWN:
                self.add_run_tags(run_id, {RUN_FAILURE_REASON_TAG: failure_reason.value})

    def _row_to_run(self, row: dict) -> DagsterRun:
        run = deserialize_value(row["run_body"], DagsterRun)
        status = DagsterRunStatus(row["status"])
        # NOTE: the status column is more trustworthy than the status in the run body, since concurrent
        # writes (e.g.  handle_run_event and add_tags) can cause the status in the body to be out of
        # overriden with an old value.
        return run.with_status(status)

    def _rows_to_runs(self, rows: Iterable[dict]) -> Sequence[DagsterRun]:
        return list(map(self._row_to_run, rows))

    def _add_cursor_limit_to_query(
        self,
        query: SqlAlchemyQuery,
        cursor: Optional[str],
        limit: Optional[int],
        order_by: Optional[str],
        ascending: Optional[bool],
    ) -> SqlAlchemyQuery:
        """Helper function to deal with cursor/limit pagination args."""
        if cursor:
            cursor_query = db_select([RunsTable.c.id]).where(RunsTable.c.run_id == cursor)
            if ascending:
                query = query.where(RunsTable.c.id > db_scalar_subquery(cursor_query))
            else:
                query = query.where(RunsTable.c.id < db_scalar_subquery(cursor_query))

        if limit:
            query = query.limit(limit)

        sorting_column = getattr(RunsTable.c, order_by) if order_by else RunsTable.c.id
        direction = db.asc if ascending else db.desc
        query = query.order_by(direction(sorting_column))

        return query

    def _add_filters_to_table(self, table: db.Table, filters: RunsFilter) -> db.Table:
        if filters.tags:
            table = self._apply_tags_table_filters(table, filters.tags)

        return table

    def _add_filters_to_query(self, query: SqlAlchemyQuery, filters: RunsFilter) -> SqlAlchemyQuery:
        check.inst_param(filters, "filters", RunsFilter)

        if filters.run_ids:
            query = query.where(RunsTable.c.run_id.in_(filters.run_ids))

        if filters.job_name:
            query = query.where(RunsTable.c.pipeline_name == filters.job_name)

        if filters.statuses:
            query = query.where(
                RunsTable.c.status.in_([status.value for status in filters.statuses])
            )

        if filters.snapshot_id:
            query = query.where(RunsTable.c.snapshot_id == filters.snapshot_id)

        if filters.updated_after:
            query = query.where(
                RunsTable.c.update_timestamp > filters.updated_after.replace(tzinfo=None)
            )

        if filters.updated_before:
            query = query.where(
                RunsTable.c.update_timestamp < filters.updated_before.replace(tzinfo=None)
            )

        if filters.created_after:
            query = query.where(
                RunsTable.c.create_timestamp > filters.created_after.replace(tzinfo=None)
            )

        if filters.created_before:
            query = query.where(
                RunsTable.c.create_timestamp < filters.created_before.replace(tzinfo=None)
            )

        if filters.exclude_subruns:
            if self.has_built_index(RUN_BACKFILL_ID):
                query = query.where(RunsTable.c.backfill_id.is_(None))
            else:
                runs_in_backfills = db_select([RunTagsTable.c.run_id]).where(
                    RunTagsTable.c.key == BACKFILL_ID_TAG
                )
                query = query.where(RunsTable.c.run_id.notin_(runs_in_backfills))

        return query

    def _runs_query(
        self,
        filters: Optional[RunsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        columns: Optional[Sequence[str]] = None,
        order_by: Optional[str] = None,
        ascending: bool = False,
        bucket_by: Optional[Union[JobBucket, TagBucket]] = None,
    ) -> SqlAlchemyQuery:
        filters = check.opt_inst_param(filters, "filters", RunsFilter, default=RunsFilter())
        check.opt_str_param(cursor, "cursor")
        check.opt_int_param(limit, "limit")
        check.opt_sequence_param(columns, "columns")
        check.opt_str_param(order_by, "order_by")
        check.opt_bool_param(ascending, "ascending")

        if columns is None:
            columns = ["run_body", "status"]

        table = self._add_filters_to_table(RunsTable, filters)
        base_query = db_select([getattr(RunsTable.c, column) for column in columns]).select_from(
            table
        )
        base_query = self._add_filters_to_query(base_query, filters)
        return self._add_cursor_limit_to_query(base_query, cursor, limit, order_by, ascending)

    def _apply_tags_table_filters(
        self, table: db.Table, tags: Mapping[str, Union[str, Sequence[str]]]
    ) -> SqlAlchemyQuery:
        """Efficient query pattern for filtering by multiple tags."""
        for i, (key, value) in enumerate(tags.items()):
            run_tags_alias = db.alias(RunTagsTable, f"run_tags_filter{i}")

            table = table.join(
                run_tags_alias,
                db.and_(
                    RunsTable.c.run_id == run_tags_alias.c.run_id,
                    run_tags_alias.c.key == key,
                    (run_tags_alias.c.value == value)
                    if isinstance(value, str)
                    else run_tags_alias.c.value.in_(value),
                ),
            )

        return table

    def get_runs(
        self,
        filters: Optional[RunsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        bucket_by: Optional[Union[JobBucket, TagBucket]] = None,
        ascending: bool = False,
    ) -> Sequence[DagsterRun]:
        query = self._runs_query(filters, cursor, limit, bucket_by=bucket_by, ascending=ascending)
        rows = self.fetchall(query)
        return self._rows_to_runs(rows)

    def get_run_ids(
        self,
        filters: Optional[RunsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Sequence[str]:
        query = self._runs_query(filters=filters, cursor=cursor, limit=limit, columns=["run_id"])
        rows = self.fetchall(query)
        return [row["run_id"] for row in rows]

    def get_runs_count(self, filters: Optional[RunsFilter] = None) -> int:
        subquery = db_subquery(self._runs_query(filters=filters))
        query = db_select([db.func.count().label("count")]).select_from(subquery)
        row = self.fetchone(query)
        count = row["count"] if row else 0
        return count

    def _get_run_by_id(self, run_id: str) -> Optional[DagsterRun]:
        check.str_param(run_id, "run_id")

        query = db_select([RunsTable.c.run_body, RunsTable.c.status]).where(
            RunsTable.c.run_id == run_id
        )
        rows = self.fetchall(query)
        return self._row_to_run(rows[0]) if rows else None

    def get_run_records(
        self,
        filters: Optional[RunsFilter] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
        ascending: bool = False,
        cursor: Optional[str] = None,
        bucket_by: Optional[Union[JobBucket, TagBucket]] = None,
    ) -> Sequence[RunRecord]:
        filters = check.opt_inst_param(filters, "filters", RunsFilter, default=RunsFilter())
        check.opt_int_param(limit, "limit")

        columns = ["id", "run_body", "status", "create_timestamp", "update_timestamp"]

        if self.has_run_stats_index_cols():
            columns += ["start_time", "end_time"]
        # only fetch columns we use to build RunRecord
        query = self._runs_query(
            filters=filters,
            limit=limit,
            columns=columns,
            order_by=order_by,
            ascending=ascending,
            cursor=cursor,
            bucket_by=bucket_by,
        )

        rows = self.fetchall(query)
        return [
            RunRecord(
                storage_id=check.int_param(row["id"], "id"),
                dagster_run=self._row_to_run(row),
                create_timestamp=utc_datetime_from_naive(
                    check.inst(row["create_timestamp"], datetime)
                ),
                update_timestamp=utc_datetime_from_naive(
                    check.inst(row["update_timestamp"], datetime)
                ),
                start_time=(
                    check.opt_inst(row["start_time"], float) if "start_time" in row else None
                ),
                end_time=check.opt_inst(row["end_time"], float) if "end_time" in row else None,
            )
            for row in rows
        ]

    def get_run_tags(
        self,
        tag_keys: Sequence[str],
        value_prefix: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Sequence[tuple[str, set[str]]]:
        result = defaultdict(set)
        query = (
            db_select([RunTagsTable.c.key, RunTagsTable.c.value])
            .distinct()
            .order_by(RunTagsTable.c.key, RunTagsTable.c.value)
            .where(RunTagsTable.c.key.in_(tag_keys))
        )
        if value_prefix:
            query = query.where(RunTagsTable.c.value.startswith(value_prefix))
        if limit:
            query = query.limit(limit)
        rows = self.fetchall(query)
        for r in rows:
            result[r["key"]].add(r["value"])
        return sorted(list([(k, v) for k, v in result.items()]), key=lambda x: x[0])

    def get_run_tag_keys(self) -> Sequence[str]:
        query = db_select([RunTagsTable.c.key]).distinct().order_by(RunTagsTable.c.key)
        rows = self.fetchall(query)
        return sorted([r["key"] for r in rows])

    def add_run_tags(self, run_id: str, new_tags: Mapping[str, str]) -> None:
        check.str_param(run_id, "run_id")
        check.mapping_param(new_tags, "new_tags", key_type=str, value_type=str)

        run = self._get_run_by_id(run_id)
        if not run:
            raise DagsterRunNotFoundError(
                f"Run {run_id} was not found in instance.", invalid_run_id=run_id
            )
        current_tags = run.tags if run.tags else {}

        all_tags = merge_dicts(current_tags, new_tags)
        partition = all_tags.get(PARTITION_NAME_TAG)
        partition_set = all_tags.get(PARTITION_SET_TAG)

        with self.connect() as conn:
            conn.execute(
                RunsTable.update()
                .where(RunsTable.c.run_id == run_id)
                .values(
                    run_body=serialize_value(run.with_tags(merge_dicts(current_tags, new_tags))),
                    partition=partition,
                    partition_set=partition_set,
                    update_timestamp=get_current_datetime(),
                )
            )

            current_tags_set = set(current_tags.keys())
            new_tags_set = set(new_tags.keys())

            existing_tags = current_tags_set & new_tags_set
            added_tags = new_tags_set.difference(existing_tags)

            for tag in existing_tags:
                conn.execute(
                    RunTagsTable.update()
                    .where(db.and_(RunTagsTable.c.run_id == run_id, RunTagsTable.c.key == tag))
                    .values(value=new_tags[tag])
                )

            if added_tags:
                conn.execute(
                    RunTagsTable.insert(),
                    [dict(run_id=run_id, key=tag, value=new_tags[tag]) for tag in added_tags],
                )

    def get_run_group(self, run_id: str) -> tuple[str, Sequence[DagsterRun]]:
        check.str_param(run_id, "run_id")
        dagster_run = self._get_run_by_id(run_id)
        if not dagster_run:
            raise DagsterRunNotFoundError(
                f"Run {run_id} was not found in instance.", invalid_run_id=run_id
            )

        # find root_run
        root_run_id = dagster_run.root_run_id if dagster_run.root_run_id else dagster_run.run_id
        root_run = self._get_run_by_id(root_run_id)
        if not root_run:
            raise DagsterRunNotFoundError(
                f"Run id {root_run_id} set as root run id for run {run_id} was not found in"
                " instance.",
                invalid_run_id=root_run_id,
            )

        # root_run_id to run_id 1:1 mapping
        # https://github.com/dagster-io/dagster/issues/2495
        # Note: we currently use tags to persist the run group info
        root_to_run = db_subquery(
            db_select(
                [RunTagsTable.c.value.label("root_run_id"), RunTagsTable.c.run_id.label("run_id")]
            ).where(
                db.and_(RunTagsTable.c.key == ROOT_RUN_ID_TAG, RunTagsTable.c.value == root_run_id)
            ),
            "root_to_run",
        )
        # get run group
        run_group_query = db_select([RunsTable.c.run_body, RunsTable.c.status]).select_from(
            root_to_run.join(
                RunsTable,
                root_to_run.c.run_id == RunsTable.c.run_id,
                isouter=True,
            )
        )

        res = self.fetchall(run_group_query)
        run_group = self._rows_to_runs(res)

        return (root_run_id, [root_run, *run_group])

    def has_run(self, run_id: str) -> bool:
        check.str_param(run_id, "run_id")
        return bool(self._get_run_by_id(run_id))

    def delete_run(self, run_id: str) -> None:
        check.str_param(run_id, "run_id")
        query = db.delete(RunsTable).where(RunsTable.c.run_id == run_id)
        with self.connect() as conn:
            conn.execute(query)

    def has_job_snapshot(self, job_snapshot_id: str) -> bool:
        check.str_param(job_snapshot_id, "job_snapshot_id")
        return self._has_snapshot_id(job_snapshot_id)

    def add_job_snapshot(self, job_snapshot: JobSnap) -> str:
        check.inst_param(job_snapshot, "job_snapshot", JobSnap)

        snapshot_id = job_snapshot.snapshot_id

        return self._add_snapshot(
            snapshot_id=snapshot_id,
            snapshot_obj=job_snapshot,
            snapshot_type=SnapshotType.PIPELINE,
        )

    def get_job_snapshot(self, job_snapshot_id: str) -> JobSnap:
        check.str_param(job_snapshot_id, "job_snapshot_id")
        return self._get_snapshot(job_snapshot_id)  # type: ignore  # (allowed to return None?)

    def has_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> bool:
        check.str_param(execution_plan_snapshot_id, "execution_plan_snapshot_id")
        return bool(self.get_execution_plan_snapshot(execution_plan_snapshot_id))

    def add_execution_plan_snapshot(
        self,
        execution_plan_snapshot: ExecutionPlanSnapshot,
    ) -> str:
        check.inst_param(execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot)

        snapshot_id = create_execution_plan_snapshot_id(execution_plan_snapshot)

        return self._add_snapshot(
            snapshot_id=snapshot_id,
            snapshot_obj=execution_plan_snapshot,
            snapshot_type=SnapshotType.EXECUTION_PLAN,
        )

    def get_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> ExecutionPlanSnapshot:
        check.str_param(execution_plan_snapshot_id, "execution_plan_snapshot_id")
        return self._get_snapshot(execution_plan_snapshot_id)  # type: ignore  # (allowed to return None?)

    def _add_snapshot(self, snapshot_id: str, snapshot_obj, snapshot_type: SnapshotType) -> str:
        check.str_param(snapshot_id, "snapshot_id")
        check.not_none_param(snapshot_obj, "snapshot_obj")
        check.inst_param(snapshot_type, "snapshot_type", SnapshotType)

        with self.connect() as conn:
            snapshot_insert = SnapshotsTable.insert().values(
                snapshot_id=snapshot_id,
                snapshot_body=zlib.compress(serialize_value(snapshot_obj).encode("utf-8")),
                snapshot_type=snapshot_type.value,
            )
            try:
                conn.execute(snapshot_insert)
            except db_exc.IntegrityError:
                # on_conflict_do_nothing equivalent
                pass

            return snapshot_id

    def get_run_storage_id(self) -> str:
        query = db_select([InstanceInfo.c.run_storage_id])
        row = self.fetchone(query)
        if not row:
            run_storage_id = str(uuid.uuid4())
            with self.connect() as conn:
                conn.execute(InstanceInfo.insert().values(run_storage_id=run_storage_id))
            return run_storage_id
        else:
            return row["run_storage_id"]

    def _has_snapshot_id(self, snapshot_id: str) -> bool:
        query = db_select([SnapshotsTable.c.snapshot_id]).where(
            SnapshotsTable.c.snapshot_id == snapshot_id
        )

        row = self.fetchone(query)

        return bool(row)

    def _get_snapshot(self, snapshot_id: str) -> Optional[JobSnap]:
        query = db_select([SnapshotsTable.c.snapshot_body]).where(
            SnapshotsTable.c.snapshot_id == snapshot_id
        )

        row = self.fetchone(query)

        return (
            defensively_unpack_execution_plan_snapshot_query(logging, [row["snapshot_body"]])  # type: ignore
            if row
            else None
        )

    def get_run_partition_data(self, runs_filter: RunsFilter) -> Sequence[RunPartitionData]:
        if self.has_built_index(RUN_PARTITIONS) and self.has_run_stats_index_cols():
            query = self._runs_query(
                filters=runs_filter,
                columns=["run_id", "status", "start_time", "end_time", "partition"],
            )
            rows = self.fetchall(query)

            # dedup by partition
            _partition_data_by_partition = {}
            for row in rows:
                if not row["partition"] or row["partition"] in _partition_data_by_partition:
                    continue

                _partition_data_by_partition[row["partition"]] = RunPartitionData(
                    run_id=row["run_id"],
                    partition=row["partition"],
                    status=DagsterRunStatus[row["status"]],
                    start_time=row["start_time"],
                    end_time=row["end_time"],
                )

            return list(_partition_data_by_partition.values())
        else:
            query = self._runs_query(filters=runs_filter)
            rows = self.fetchall(query)
            _partition_data_by_partition = {}
            for row in rows:
                run = self._row_to_run(row)
                partition = run.tags.get(PARTITION_NAME_TAG)
                if not partition or partition in _partition_data_by_partition:
                    continue

                _partition_data_by_partition[partition] = RunPartitionData(
                    run_id=run.run_id,
                    partition=partition,
                    status=run.status,
                    start_time=None,
                    end_time=None,
                )

            return list(_partition_data_by_partition.values())

    def _get_partition_runs(
        self, partition_set_name: str, partition_name: str
    ) -> Sequence[DagsterRun]:
        # utility method to help test reads off of the partition column
        if not self.has_built_index(RUN_PARTITIONS):
            # query by tags
            return self.get_runs(
                filters=RunsFilter(
                    tags={
                        PARTITION_SET_TAG: partition_set_name,
                        PARTITION_NAME_TAG: partition_name,
                    }
                )
            )
        else:
            query = (
                self._runs_query()
                .where(RunsTable.c.partition == partition_name)
                .where(RunsTable.c.partition_set == partition_set_name)
            )
            rows = self.fetchall(query)
            return self._rows_to_runs(rows)

    # Tracking data migrations over secondary indexes

    def _execute_data_migrations(
        self,
        migrations: Mapping[str, Callable[[], MigrationFn]],
        print_fn: Optional[PrintFn] = None,
        force_rebuild_all: bool = False,
    ) -> None:
        for migration_name, migration_fn in migrations.items():
            if self.has_built_index(migration_name):
                if not force_rebuild_all:
                    if print_fn:
                        print_fn(f"Skipping already applied data migration: {migration_name}")
                    continue
            if print_fn:
                print_fn(f"Starting data migration: {migration_name}")
            migration_fn()(self, print_fn)
            self.mark_index_built(migration_name)
            if print_fn:
                print_fn(f"Finished data migration: {migration_name}")

    def migrate(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        self._execute_data_migrations(REQUIRED_DATA_MIGRATIONS, print_fn, force_rebuild_all)

    def optimize(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        self._execute_data_migrations(OPTIONAL_DATA_MIGRATIONS, print_fn, force_rebuild_all)

    def has_built_index(self, migration_name: str) -> bool:
        query = (
            db_select([1])
            .where(SecondaryIndexMigrationTable.c.name == migration_name)
            .where(SecondaryIndexMigrationTable.c.migration_completed != None)  # noqa: E711
            .limit(1)
        )
        results = self.fetchall(query)

        return len(results) > 0

    def mark_index_built(self, migration_name: str) -> None:
        query = SecondaryIndexMigrationTable.insert().values(
            name=migration_name,
            migration_completed=datetime.now(),
        )
        with self.connect() as conn:
            try:
                conn.execute(query)
            except db_exc.IntegrityError:
                conn.execute(
                    SecondaryIndexMigrationTable.update()
                    .where(SecondaryIndexMigrationTable.c.name == migration_name)
                    .values(migration_completed=datetime.now())
                )

    # Checking for migrations

    def has_run_stats_index_cols(self) -> bool:
        with self.connect() as conn:
            column_names = [x.get("name") for x in db.inspect(conn).get_columns(RunsTable.name)]
            return "start_time" in column_names and "end_time" in column_names

    def has_bulk_actions_selector_cols(self) -> bool:
        with self.connect() as conn:
            column_names = [
                x.get("name") for x in db.inspect(conn).get_columns(BulkActionsTable.name)
            ]
            return "selector_id" in column_names

    def has_backfill_id_column(self) -> bool:
        with self.connect() as conn:
            column_names = [x.get("name") for x in db.inspect(conn).get_columns(RunsTable.name)]
            return "backfill_id" in column_names

    def has_bulk_action_job_name_column(self) -> bool:
        with self.connect() as conn:
            column_names = [
                x.get("name") for x in db.inspect(conn).get_columns(BulkActionsTable.name)
            ]
            return "job_name" in column_names

    def has_backfill_tags_table(self) -> bool:
        with self.connect() as conn:
            return BackfillTagsTable.name in db.inspect(conn).get_table_names()

    # Daemon heartbeats

    def add_daemon_heartbeat(self, daemon_heartbeat: DaemonHeartbeat) -> None:
        with self.connect() as conn:
            # insert, or update if already present
            try:
                conn.execute(
                    DaemonHeartbeatsTable.insert().values(
                        timestamp=datetime_from_timestamp(daemon_heartbeat.timestamp),
                        daemon_type=daemon_heartbeat.daemon_type,
                        daemon_id=daemon_heartbeat.daemon_id,
                        body=serialize_value(daemon_heartbeat),
                    )
                )
            except db_exc.IntegrityError:
                conn.execute(
                    DaemonHeartbeatsTable.update()
                    .where(DaemonHeartbeatsTable.c.daemon_type == daemon_heartbeat.daemon_type)
                    .values(
                        timestamp=datetime_from_timestamp(daemon_heartbeat.timestamp),
                        daemon_id=daemon_heartbeat.daemon_id,
                        body=serialize_value(daemon_heartbeat),
                    )
                )

    def get_daemon_heartbeats(self) -> Mapping[str, DaemonHeartbeat]:
        rows = self.fetchall(db_select([DaemonHeartbeatsTable.c.body]))
        heartbeats = []
        for row in rows:
            heartbeats.append(deserialize_value(row["body"], DaemonHeartbeat))
        return {heartbeat.daemon_type: heartbeat for heartbeat in heartbeats}

    def wipe(self) -> None:
        """Clears the run storage."""
        with self.connect() as conn:
            # https://stackoverflow.com/a/54386260/324449
            conn.execute(RunsTable.delete())
            conn.execute(RunTagsTable.delete())
            conn.execute(SnapshotsTable.delete())
            conn.execute(DaemonHeartbeatsTable.delete())
            conn.execute(BulkActionsTable.delete())

    def wipe_daemon_heartbeats(self) -> None:
        with self.connect() as conn:
            # https://stackoverflow.com/a/54386260/324449
            conn.execute(DaemonHeartbeatsTable.delete())

    def _add_backfill_filters_to_table(
        self, table: db.Table, filters: Optional[BulkActionsFilter]
    ) -> db.Table:
        if filters and filters.tags and self.has_built_index(BACKFILL_JOB_NAME_AND_TAGS):
            for i, (key, value) in enumerate(filters.tags.items()):
                backfill_tags_alias = db.alias(BackfillTagsTable, f"backfill_tags_filter{i}")

                table = table.join(
                    backfill_tags_alias,
                    db.and_(
                        BulkActionsTable.c.key == backfill_tags_alias.c.backfill_id,
                        backfill_tags_alias.c.key == key,
                        (backfill_tags_alias.c.value == value)
                        if isinstance(value, str)
                        else backfill_tags_alias.c.value.in_(value),
                    ),
                )
            return table
        return table

    def _backfills_query(self, filters: Optional[BulkActionsFilter] = None):
        query = db_select([BulkActionsTable.c.body, BulkActionsTable.c.timestamp])
        if filters and filters.tags:
            if not self.has_built_index(BACKFILL_JOB_NAME_AND_TAGS):
                # if the migration was run, we added the query for tags filtering in _add_backfill_filters_to_table
                # BackfillTags table has not been built. However, all tags that are on a backfill are
                # applied to the runs the backfill launches. So we can query for runs that match the tags and
                # are also part of a backfill to find the backfills that match the tags.

                backfills_with_tags_query = db_select([RunTagsTable.c.value]).where(
                    RunTagsTable.c.key == BACKFILL_ID_TAG
                )

                for i, (key, value) in enumerate(filters.tags.items()):
                    run_tags_alias = db.alias(RunTagsTable, f"run_tags_filter{i}")
                    backfills_with_tags_query = backfills_with_tags_query.where(
                        db.and_(
                            RunTagsTable.c.run_id == run_tags_alias.c.run_id,
                            run_tags_alias.c.key == key,
                            (run_tags_alias.c.value == value)
                            if isinstance(value, str)
                            else run_tags_alias.c.value.in_(value),
                        ),
                    )

                query = query.where(
                    BulkActionsTable.c.key.in_(db_subquery(backfills_with_tags_query))
                )

        if filters and filters.job_name:
            if self.has_built_index(BACKFILL_JOB_NAME_AND_TAGS):
                query = query.where(BulkActionsTable.c.job_name == filters.job_name)
            else:
                run_tags_table = RunTagsTable

                runs_in_backfill_with_job_name = run_tags_table.join(
                    RunsTable,
                    db.and_(
                        RunTagsTable.c.run_id == RunsTable.c.run_id,
                        RunTagsTable.c.key == BACKFILL_ID_TAG,
                        RunsTable.c.pipeline_name == filters.job_name,
                    ),
                )

                backfills_with_job_name_query = db_select([RunTagsTable.c.value]).select_from(
                    runs_in_backfill_with_job_name
                )
                query = query.where(
                    BulkActionsTable.c.key.in_(db_subquery(backfills_with_job_name_query))
                )
        if filters and filters.statuses:
            query = query.where(
                BulkActionsTable.c.status.in_([status.value for status in filters.statuses])
            )
        if filters and filters.created_after:
            query = query.where(BulkActionsTable.c.timestamp > filters.created_after)
        if filters and filters.created_before:
            query = query.where(BulkActionsTable.c.timestamp < filters.created_before)
        if filters and filters.backfill_ids:
            query = query.where(BulkActionsTable.c.key.in_(filters.backfill_ids))
        return query

    def _add_cursor_limit_to_backfills_query(
        self, query, cursor: Optional[str] = None, limit: Optional[int] = None
    ):
        if limit:
            query = query.limit(limit)
        if cursor:
            cursor_query = db_select([BulkActionsTable.c.id]).where(
                BulkActionsTable.c.key == cursor
            )
            query = query.where(BulkActionsTable.c.id < cursor_query)

        return query

    def _apply_backfill_tags_filter_to_results(
        self, backfills: Sequence[PartitionBackfill], tags: Mapping[str, Union[str, Sequence[str]]]
    ) -> Sequence[PartitionBackfill]:
        if not tags:
            return backfills

        def _matches_backfill(
            backfill: PartitionBackfill, tags: Mapping[str, Union[str, Sequence[str]]]
        ) -> bool:
            for key, value in tags.items():
                if isinstance(value, str):
                    if backfill.tags.get(key) != value:
                        return False
                elif backfill.tags.get(key) not in value:
                    return False
            return True

        return [backfill for backfill in backfills if _matches_backfill(backfill, tags)]

    def get_backfills(
        self,
        filters: Optional[BulkActionsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        status: Optional[BulkActionStatus] = None,
    ) -> Sequence[PartitionBackfill]:
        check.opt_inst_param(status, "status", BulkActionStatus)
        query = db_select([BulkActionsTable.c.body, BulkActionsTable.c.timestamp])
        if status and filters:
            raise DagsterInvariantViolationError(
                "Cannot provide status and filters to get_backfills. Please use filters rather than status."
            )

        if status is not None:
            filters = BulkActionsFilter(statuses=[status])

        table = self._add_backfill_filters_to_table(BulkActionsTable, filters)
        query = self._backfills_query(filters=filters).select_from(table)
        query = self._add_cursor_limit_to_backfills_query(query, cursor=cursor, limit=limit)
        query = query.order_by(BulkActionsTable.c.id.desc())
        rows = self.fetchall(query)
        backfill_candidates = deserialize_values((row["body"] for row in rows), PartitionBackfill)

        if filters and filters.tags and not self.has_built_index(BACKFILL_JOB_NAME_AND_TAGS):
            # if we are still using the run tags table to get backfills by tag, we need to do an additional check.
            # runs can have more tags than the backfill that launched them. Since we filtered tags by
            # querying for runs with those tags, we need to do an additional check that the backfills
            # also have the requested tags
            backfill_candidates = self._apply_backfill_tags_filter_to_results(
                backfill_candidates, filters.tags
            )
        return backfill_candidates

    def get_backfills_count(self, filters: Optional[BulkActionsFilter] = None) -> int:
        check.opt_inst_param(filters, "filters", BulkActionsFilter)
        if filters and filters.tags:
            # runs can have more tags than the backfill that launched them. Since we filtered tags by
            # querying for runs with those tags, we need to do an additional check that the backfills
            # also have the requested tags. This requires fetching the backfills from the db and filtering them
            query = self._backfills_query(filters=filters)
            rows = self.fetchall(query)
            backfill_candidates = deserialize_values(
                (row["body"] for row in rows), PartitionBackfill
            )
            return len(
                self._apply_backfill_tags_filter_to_results(backfill_candidates, filters.tags)
            )

        subquery = db_subquery(self._backfills_query(filters=filters))
        query = db_select([db.func.count().label("count")]).select_from(subquery)
        row = self.fetchone(query)
        count = row["count"] if row else 0
        return count

    def get_backfill(self, backfill_id: str) -> Optional[PartitionBackfill]:
        check.str_param(backfill_id, "backfill_id")
        query = db_select([BulkActionsTable.c.body]).where(BulkActionsTable.c.key == backfill_id)
        row = self.fetchone(query)
        return deserialize_value(row["body"], PartitionBackfill) if row else None

    def add_backfill(self, partition_backfill: PartitionBackfill) -> None:
        check.inst_param(partition_backfill, "partition_backfill", PartitionBackfill)
        values: dict[str, Any] = dict(
            key=partition_backfill.backfill_id,
            status=partition_backfill.status.value,
            timestamp=datetime_from_timestamp(partition_backfill.backfill_timestamp),
            body=serialize_value(cast(NamedTuple, partition_backfill)),
        )

        if self.has_bulk_actions_selector_cols():
            values["selector_id"] = partition_backfill.selector_id
            values["action_type"] = partition_backfill.bulk_action_type.value

        if self.has_bulk_action_job_name_column():
            values["job_name"] = partition_backfill.job_name

        with self.connect() as conn:
            conn.execute(BulkActionsTable.insert().values(**values))
            if self.has_backfill_tags_table():
                tags_to_insert = partition_backfill.tags
                if len(tags_to_insert.items()) > 0:
                    conn.execute(
                        BackfillTagsTable.insert(),
                        [
                            dict(
                                backfill_id=partition_backfill.backfill_id,
                                key=k,
                                value=v,
                            )
                            for k, v in tags_to_insert.items()
                        ],
                    )

    def update_backfill(self, partition_backfill: PartitionBackfill) -> None:
        check.inst_param(partition_backfill, "partition_backfill", PartitionBackfill)
        backfill_id = partition_backfill.backfill_id
        if not self.get_backfill(backfill_id):
            raise DagsterInvariantViolationError(
                f"Backfill {backfill_id} is not present in storage"
            )
        with self.connect() as conn:
            conn.execute(
                BulkActionsTable.update()
                .where(BulkActionsTable.c.key == backfill_id)
                .values(
                    status=partition_backfill.status.value,
                    body=serialize_value(partition_backfill),
                )
            )

    def get_cursor_values(self, keys: set[str]) -> Mapping[str, str]:
        check.set_param(keys, "keys", of_type=str)

        rows = self.fetchall(
            db_select([KeyValueStoreTable.c.key, KeyValueStoreTable.c.value]).where(
                KeyValueStoreTable.c.key.in_(keys)
            ),
        )
        return {row["key"]: row["value"] for row in rows}

    def set_cursor_values(self, pairs: Mapping[str, str]) -> None:
        check.mapping_param(pairs, "pairs", key_type=str, value_type=str)
        db_values = [{"key": k, "value": v} for k, v in pairs.items()]

        with self.connect() as conn:
            try:
                conn.execute(KeyValueStoreTable.insert().values(db_values))
            except db_exc.IntegrityError:
                conn.execute(
                    KeyValueStoreTable.update()
                    .where(KeyValueStoreTable.c.key.in_(pairs.keys()))
                    .values(value=db.sql.case(pairs, value=KeyValueStoreTable.c.key))
                )

    # Migrating run history
    def replace_job_origin(self, run: DagsterRun, job_origin: RemoteJobOrigin) -> None:
        new_label = job_origin.repository_origin.get_label()
        with self.connect() as conn:
            conn.execute(
                RunsTable.update()
                .where(RunsTable.c.run_id == run.run_id)
                .values(
                    run_body=serialize_value(
                        run.with_job_origin(job_origin).with_tags(
                            {**run.tags, REPOSITORY_LABEL_TAG: new_label}
                        )
                    ),
                )
            )
            conn.execute(
                RunTagsTable.update()
                .where(RunTagsTable.c.run_id == run.run_id)
                .where(RunTagsTable.c.key == REPOSITORY_LABEL_TAG)
                .values(value=new_label)
            )


GET_PIPELINE_SNAPSHOT_QUERY_ID = "get-pipeline-snapshot"


def defensively_unpack_execution_plan_snapshot_query(
    logger: logging.Logger, row: Sequence[Any]
) -> Optional[Union[ExecutionPlanSnapshot, JobSnap]]:
    # minimal checking here because sqlalchemy returns a different type based on what version of
    # SqlAlchemy you are using

    def _warn(msg: str) -> None:
        logger.warning(f"get-pipeline-snapshot: {msg}")

    if not isinstance(row[0], bytes):
        _warn("First entry in row is not a binary type.")
        return None

    try:
        uncompressed_bytes = zlib.decompress(row[0])
    except zlib.error:
        _warn("Could not decompress bytes stored in snapshot table.")
        return None

    try:
        decoded_str = uncompressed_bytes.decode("utf-8")
    except UnicodeDecodeError:
        _warn("Could not unicode decode decompressed bytes stored in snapshot table.")
        return None

    try:
        return deserialize_value(decoded_str, (ExecutionPlanSnapshot, JobSnap))
    except JSONDecodeError:
        _warn("Could not parse json in snapshot table.")
        return None



from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.resource_requirement import ensure_requirements_satisfied
from dagster._core.execution.context.hook import BoundHookContext, UnboundHookContext

if TYPE_CHECKING:
    from dagster._core.definitions.hook_definition import HookDefinition
    from dagster._core.events import DagsterEvent


def hook_invocation_result(
    hook_def: "HookDefinition",
    hook_context: Optional[UnboundHookContext],
    event_list: Optional[Sequence["DagsterEvent"]] = None,
):
    if not hook_context:
        hook_context = UnboundHookContext(
            resources={}, op=None, run_id=None, job_name=None, op_exception=None, instance=None
        )

    # Validate that all required resources are provided in the context

    ensure_requirements_satisfied(
        hook_context._resource_defs,  # noqa: SLF001
        list(hook_def.get_resource_requirements(attached_to=None)),
    )

    bound_context = BoundHookContext(
        hook_def=hook_def,
        resources=hook_context.resources,
        log_manager=hook_context.log,
        op=hook_context._op,  # noqa: SLF001
        run_id=hook_context._run_id,  # noqa: SLF001
        job_name=hook_context._job_name,  # noqa: SLF001
        op_exception=hook_context._op_exception,  # noqa: SLF001
        instance=hook_context._instance,  # noqa: SLF001
    )

    decorated_fn = check.not_none(hook_def.decorated_fn)

    return (
        decorated_fn(bound_context, event_list)
        if event_list is not None
        else decorated_fn(bound_context)
    )



"""add runs table backfill_id index

Revision ID: 1aca709bba64
Revises: 63d7a8ec641a
Create Date: 2024-10-23 17:36:39.876876

"""

from alembic import op
from dagster._core.storage.migration.utils import has_index, has_table

# revision identifiers, used by Alembic.
revision = "1aca709bba64"
down_revision = "63d7a8ec641a"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("runs"):
        return

    if not has_index("runs", "idx_runs_by_backfill_id"):
        op.create_index(
            "idx_runs_by_backfill_id",
            "runs",
            ["backfill_id", "id"],
            unique=False,
            postgresql_concurrently=True,
            mysql_length={"backfill_id": 255},
        )


def downgrade():
    if not has_table("runs"):
        return

    if has_index("runs", "idx_runs_by_backfill_id"):
        op.drop_index(
            "idx_runs_by_backfill_id",
            "runs",
            postgresql_concurrently=True,
        )



"""add column mode.

Revision ID: ddcc6d7244c6
Revises: 45fa602c43dc
Create Date: 2021-07-06 13:56:08.987201

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import inspect

# revision identifiers, used by Alembic.
revision = "ddcc6d7244c6"
down_revision = "45fa602c43dc"
branch_labels = None
depends_on = None


def upgrade():
    inspector = inspect(op.get_bind())
    has_tables = inspector.get_table_names()

    if "runs" in has_tables:
        columns = [x.get("name") for x in inspector.get_columns("runs")]
        with op.batch_alter_table("runs") as batch_op:
            if "mode" not in columns:
                batch_op.add_column(sa.Column("mode", sa.Text))


def downgrade():
    inspector = inspect(op.get_bind())
    has_tables = inspector.get_table_names()
    if "runs" in has_tables:
        columns = [x.get("name") for x in inspector.get_columns("runs")]

        with op.batch_alter_table("runs") as batch_op:
            if "mode" in columns:
                batch_op.drop_column("mode")



from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, AbstractSet, Any, Callable, Optional, cast  # noqa: UP035

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import PublicAttr, beta_param, deprecated, deprecated_param, public
from dagster._core.decorator_utils import get_function_params
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.data_version import (
    DATA_VERSION_TAG,
    DataVersion,
    DataVersionsByPartition,
)
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.events import AssetKey, AssetObservation, CoercibleToAssetKey, Output
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import (
    ArbitraryMetadataMapping,
    MetadataMapping,
    normalize_metadata,
)
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.resource_requirement import (
    ResourceAddable,
    ResourceKeyRequirement,
    ResourceRequirement,
    SourceAssetIOManagerRequirement,
    ensure_requirements_satisfied,
    get_resource_key_conflicts,
)
from dagster._core.definitions.result import ObserveResult
from dagster._core.definitions.utils import (
    DEFAULT_GROUP_NAME,
    DEFAULT_IO_MANAGER_KEY,
    normalize_group_name,
)
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvalidObservationError,
    DagsterInvariantViolationError,
)
from dagster._utils.internal_init import IHasInternalInit
from dagster._utils.tags import normalize_tags

if TYPE_CHECKING:
    from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
from dagster._core.storage.io_manager import IOManagerDefinition
from dagster._utils.merger import merge_dicts
from dagster._utils.warnings import disable_dagster_warnings

# Going with this catch-all for the time-being to permit pythonic resources
SourceAssetObserveFunction: TypeAlias = Callable[..., Any]

# This is a private key that is attached to the Output emitted from a source asset observation
# function and used to prevent observations from being auto-generated from it. This is a workaround
# because we cannot currently auto-convert the observation function to use `ObserveResult`. It can
# be removed when that conversion is completed.
SYSTEM_METADATA_KEY_SOURCE_ASSET_OBSERVATION = "__source_asset_observation__"


def wrap_source_asset_observe_fn_in_op_compute_fn(
    source_asset: "SourceAsset",
) -> "DecoratedOpFunction":
    from dagster._core.definitions.decorators.op_decorator import (
        DecoratedOpFunction,
        is_context_provided,
    )
    from dagster._core.execution.context.compute import OpExecutionContext

    check.not_none(source_asset.observe_fn, "Must be an observable source asset")
    assert source_asset.observe_fn  # for type checker

    observe_fn = source_asset.observe_fn

    observe_fn_has_context = is_context_provided(get_function_params(observe_fn))

    def fn(context: OpExecutionContext) -> Output[None]:
        resource_kwarg_keys = [param.name for param in get_resource_args(observe_fn)]
        resource_kwargs = {
            key: context.resources.original_resource_dict.get(key) for key in resource_kwarg_keys
        }
        observe_fn_return_value = (
            observe_fn(context, **resource_kwargs)
            if observe_fn_has_context
            else observe_fn(**resource_kwargs)
        )

        if isinstance(observe_fn_return_value, (DataVersion, ObserveResult)):
            if source_asset.partitions_def is not None:
                raise DagsterInvalidObservationError(
                    f"{source_asset.key} is partitioned. Returning `{observe_fn_return_value.__class__}` not supported"
                    " for partitioned assets. Return `DataVersionsByPartition` instead."
                )

            if isinstance(observe_fn_return_value, ObserveResult):
                data_version = observe_fn_return_value.data_version
                metadata = observe_fn_return_value.metadata
                extra_tags = observe_fn_return_value.tags or {}

                if (
                    observe_fn_return_value.asset_key is not None
                    and observe_fn_return_value.asset_key != source_asset.key
                ):
                    raise DagsterInvariantViolationError(
                        f"Asset key {observe_fn_return_value.asset_key.to_user_string()} not found in AssetsDefinition"
                    )
            else:  # DataVersion
                data_version = observe_fn_return_value
                metadata = {}
                extra_tags = {}

            tags = {
                **({DATA_VERSION_TAG: data_version.value} if data_version is not None else {}),
                **extra_tags,
            }

            context.log_event(
                AssetObservation(
                    asset_key=source_asset.key,
                    tags=tags,
                    metadata=metadata,
                )
            )

        elif isinstance(observe_fn_return_value, DataVersionsByPartition):
            if source_asset.partitions_def is None:
                raise DagsterInvalidObservationError(
                    f"{source_asset.key} is not partitioned, so its observe function should return"
                    " a DataVersion, not a DataVersionsByPartition"
                )

            for (
                partition_key,
                data_version,
            ) in observe_fn_return_value.data_versions_by_partition.items():
                context.log_event(
                    AssetObservation(
                        asset_key=source_asset.key,
                        tags={DATA_VERSION_TAG: data_version.value},
                        partition=partition_key,
                    )
                )
        else:
            raise DagsterInvalidObservationError(
                f"Observe function for {source_asset.key} must return a DataVersion or"
                " DataVersionsByPartition, but returned a value of type"
                f" {type(observe_fn_return_value)}"
            )
        return Output(None, metadata={SYSTEM_METADATA_KEY_SOURCE_ASSET_OBSERVATION: True})

    return DecoratedOpFunction(fn)


@beta_param(param="resource_defs")
@beta_param(param="io_manager_def")
@deprecated_param(param="freshness_policy", breaking_version="1.11.0")
@deprecated(
    breaking_version="2.0.0",
    additional_warn_text="Use AssetSpec instead. If using the SourceAsset io_manager_key property, "
    "use AssetSpec(...).with_io_manager_key(...).",
)
class SourceAsset(ResourceAddable, IHasInternalInit):
    """A SourceAsset represents an asset that will be loaded by (but not updated by) Dagster.

    Args:
        key (Union[AssetKey, Sequence[str], str]): The key of the asset.
        metadata (Mapping[str, MetadataValue]): Metadata associated with the asset.
        io_manager_key (Optional[str]): The key for the IOManager that will be used to load the contents of
            the asset when it's used as an input to other assets inside a job.
        io_manager_def (Optional[IOManagerDefinition]): (Beta) The definition of the IOManager that will be used to load the contents of
            the asset when it's used as an input to other assets inside a job.
        resource_defs (Optional[Mapping[str, ResourceDefinition]]): (Beta) resource definitions that may be required by the :py:class:`dagster.IOManagerDefinition` provided in the `io_manager_def` argument.
        description (Optional[str]): The description of the asset.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
        observe_fn (Optional[SourceAssetObserveFunction]) Observation function for the source asset.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that computes the asset.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        auto_observe_interval_minutes (Optional[float]): While the asset daemon is turned on, a run
            of the observation function for this asset will be launched at this interval. `observe_fn`
            must be provided.
        freshness_policy (FreshnessPolicy): A constraint telling Dagster how often this asset is intended to be updated
            with respect to its root data.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
    """

    key: PublicAttr[AssetKey]
    metadata: PublicAttr[MetadataMapping]
    raw_metadata: PublicAttr[ArbitraryMetadataMapping]
    io_manager_key: PublicAttr[Optional[str]]
    _io_manager_def: PublicAttr[Optional[IOManagerDefinition]]
    description: PublicAttr[Optional[str]]
    partitions_def: PublicAttr[Optional[PartitionsDefinition]]
    group_name: PublicAttr[str]
    resource_defs: PublicAttr[dict[str, ResourceDefinition]]
    observe_fn: PublicAttr[Optional[SourceAssetObserveFunction]]
    op_tags: Optional[Mapping[str, Any]]
    _node_def: Optional[OpDefinition]  # computed lazily
    auto_observe_interval_minutes: Optional[float]
    freshness_policy: Optional[FreshnessPolicy]
    automation_condition: Optional[AutomationCondition]
    tags: Mapping[str, str]

    def __init__(
        self,
        key: CoercibleToAssetKey,
        metadata: Optional[ArbitraryMetadataMapping] = None,
        io_manager_key: Optional[str] = None,
        io_manager_def: Optional[object] = None,
        description: Optional[str] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
        group_name: Optional[str] = None,
        resource_defs: Optional[Mapping[str, object]] = None,
        observe_fn: Optional[SourceAssetObserveFunction] = None,
        op_tags: Optional[Mapping[str, Any]] = None,
        *,
        auto_observe_interval_minutes: Optional[float] = None,
        freshness_policy: Optional[FreshnessPolicy] = None,
        automation_condition: Optional[AutomationCondition] = None,
        tags: Optional[Mapping[str, str]] = None,
        # This is currently private because it is necessary for source asset observation functions,
        # but we have not yet decided on a final API for associated one or more ops with a source
        # asset. If we were to make this public, then we would have a canonical public
        # `required_resource_keys` used for observation that might end up conflicting with a set of
        # required resource keys for a different operation.
        _required_resource_keys: Optional[AbstractSet[str]] = None,
        # Add additional fields to with_resources and with_group below
    ):
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        self.key = AssetKey.from_coercible(key)
        metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
        self.raw_metadata = metadata
        self.metadata = normalize_metadata(metadata, allow_invalid=True)
        self.tags = normalize_tags(tags or {}, strict=True)

        resource_defs_dict = dict(check.opt_mapping_param(resource_defs, "resource_defs"))
        if io_manager_def:
            if not io_manager_key:
                io_manager_key = self.key.to_python_identifier("io_manager")

            if (
                io_manager_key in resource_defs_dict
                and resource_defs_dict[io_manager_key] != io_manager_def
            ):
                raise DagsterInvalidDefinitionError(
                    f"Provided conflicting definitions for io manager key '{io_manager_key}'."
                    " Please provide only one definition per key."
                )

            resource_defs_dict[io_manager_key] = io_manager_def

        self.resource_defs = wrap_resources_for_execution(resource_defs_dict)

        self.io_manager_key = check.opt_str_param(io_manager_key, "io_manager_key")
        self.partitions_def = check.opt_inst_param(
            partitions_def, "partitions_def", PartitionsDefinition
        )
        self.group_name = normalize_group_name(group_name)
        self.description = check.opt_str_param(description, "description")
        self.observe_fn = check.opt_callable_param(observe_fn, "observe_fn")
        self.op_tags = check.opt_mapping_param(op_tags, "op_tags")
        self._required_resource_keys = check.opt_set_param(
            _required_resource_keys, "_required_resource_keys", of_type=str
        )
        self._node_def = None
        self.auto_observe_interval_minutes = check.opt_numeric_param(
            auto_observe_interval_minutes, "auto_observe_interval_minutes"
        )
        self.freshness_policy = check.opt_inst_param(
            freshness_policy, "freshness_policy", FreshnessPolicy
        )
        self.automation_condition = check.opt_inst_param(
            automation_condition, "automation_condition", AutomationCondition
        )

    @staticmethod
    def dagster_internal_init(
        *,
        key: CoercibleToAssetKey,
        metadata: Optional[ArbitraryMetadataMapping],
        io_manager_key: Optional[str],
        io_manager_def: Optional[object],
        description: Optional[str],
        partitions_def: Optional[PartitionsDefinition],
        group_name: Optional[str],
        resource_defs: Optional[Mapping[str, object]],
        observe_fn: Optional[SourceAssetObserveFunction],
        op_tags: Optional[Mapping[str, Any]],
        auto_observe_interval_minutes: Optional[float],
        freshness_policy: Optional[FreshnessPolicy],
        automation_condition: Optional[AutomationCondition],
        tags: Optional[Mapping[str, str]],
        _required_resource_keys: Optional[AbstractSet[str]],
    ) -> "SourceAsset":
        return SourceAsset(
            key=key,
            metadata=metadata,
            io_manager_key=io_manager_key,
            io_manager_def=io_manager_def,
            description=description,
            partitions_def=partitions_def,
            group_name=group_name,
            resource_defs=resource_defs,
            observe_fn=observe_fn,
            op_tags=op_tags,
            auto_observe_interval_minutes=auto_observe_interval_minutes,
            freshness_policy=freshness_policy,
            automation_condition=automation_condition,
            tags=tags,
            _required_resource_keys=_required_resource_keys,
        )

    def get_io_manager_key(self) -> str:
        return self.io_manager_key or DEFAULT_IO_MANAGER_KEY

    @property
    def io_manager_def(self) -> Optional[IOManagerDefinition]:
        io_manager_key = self.get_io_manager_key()
        return cast(
            Optional[IOManagerDefinition],
            self.resource_defs.get(io_manager_key) if io_manager_key else None,
        )

    @public
    @property
    def op(self) -> OpDefinition:
        """OpDefinition: The OpDefinition associated with the observation function of an observable
        source asset.

        Throws an error if the asset is not observable.
        """
        check.invariant(
            isinstance(self.node_def, OpDefinition),
            "The NodeDefinition for this AssetsDefinition is not of type OpDefinition.",
        )
        return cast(OpDefinition, self.node_def)

    @property
    def execution_type(self) -> AssetExecutionType:
        return (
            AssetExecutionType.OBSERVATION
            if self.is_observable
            else AssetExecutionType.UNEXECUTABLE
        )

    @property
    def is_executable(self) -> bool:
        """bool: Whether the asset is observable."""
        return self.is_observable

    @public
    @property
    def is_observable(self) -> bool:
        """bool: Whether the asset is observable."""
        return self.node_def is not None

    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        return {
            requirement.key
            for requirement in self.get_resource_requirements()
            if isinstance(requirement, ResourceKeyRequirement)
        }

    @property
    def node_def(self) -> Optional[OpDefinition]:
        """Op that generates observation metadata for a source asset."""
        if self.observe_fn is None:
            return None

        if self._node_def is None:
            self._node_def = OpDefinition(
                compute_fn=wrap_source_asset_observe_fn_in_op_compute_fn(self),
                name=self.key.to_python_identifier(),
                description=self.description,
                required_resource_keys=self._required_resource_keys,
                tags=self.op_tags,
            )
        return self._node_def

    def with_resources(self, resource_defs) -> "SourceAsset":
        from dagster._core.execution.resources_init import get_transitive_required_resource_keys

        overlapping_keys = get_resource_key_conflicts(self.resource_defs, resource_defs)
        if overlapping_keys:
            raise DagsterInvalidInvocationError(
                f"SourceAsset with key {self.key} has conflicting resource "
                "definitions with provided resources for the following keys: "
                f"{sorted(list(overlapping_keys))}. Either remove the existing "
                "resources from the asset or change the resource keys so that "
                "they don't overlap."
            )

        merged_resource_defs = merge_dicts(resource_defs, self.resource_defs)

        # Ensure top-level resource requirements are met - except for
        # io_manager, since that is a default it can be resolved later.
        ensure_requirements_satisfied(merged_resource_defs, list(self.get_resource_requirements()))

        io_manager_def = merged_resource_defs.get(self.get_io_manager_key())
        if not io_manager_def and self.get_io_manager_key() != DEFAULT_IO_MANAGER_KEY:
            raise DagsterInvalidDefinitionError(
                f"SourceAsset with asset key {self.key} requires IO manager with key"
                f" '{self.get_io_manager_key()}', but none was provided."
            )
        relevant_keys = get_transitive_required_resource_keys(
            {*self._required_resource_keys, self.get_io_manager_key()}, merged_resource_defs
        )

        relevant_resource_defs = {
            key: resource_def
            for key, resource_def in merged_resource_defs.items()
            if key in relevant_keys
        }

        io_manager_key = (
            self.get_io_manager_key()
            if self.get_io_manager_key() != DEFAULT_IO_MANAGER_KEY
            else None
        )
        with disable_dagster_warnings():
            return SourceAsset.dagster_internal_init(
                key=self.key,
                io_manager_key=io_manager_key,
                description=self.description,
                partitions_def=self.partitions_def,
                metadata=self.raw_metadata,
                io_manager_def=None,
                resource_defs=relevant_resource_defs,
                group_name=self.group_name,
                observe_fn=self.observe_fn,
                auto_observe_interval_minutes=self.auto_observe_interval_minutes,
                freshness_policy=self.freshness_policy,
                tags=self.tags,
                op_tags=self.op_tags,
                automation_condition=self.automation_condition,
                _required_resource_keys=self._required_resource_keys,
            )

    def with_attributes(
        self, group_name: Optional[str] = None, key: Optional[AssetKey] = None
    ) -> "SourceAsset":
        if (
            group_name is not None
            and self.group_name != DEFAULT_GROUP_NAME
            and self.group_name != group_name
        ):
            raise DagsterInvalidDefinitionError(
                f"Attempted to override group name to {group_name} for SourceAsset {self.key.to_user_string()}, which already has group name {self.group_name}."
                f" {self.key.to_user_string()}"
            )

        with disable_dagster_warnings():
            return SourceAsset.dagster_internal_init(
                key=key or self.key,
                metadata=self.raw_metadata,
                io_manager_key=self.io_manager_key,
                io_manager_def=self.io_manager_def,
                description=self.description,
                partitions_def=self.partitions_def,
                group_name=group_name or self.group_name,
                resource_defs=self.resource_defs,
                observe_fn=self.observe_fn,
                auto_observe_interval_minutes=self.auto_observe_interval_minutes,
                tags=self.tags,
                freshness_policy=self.freshness_policy,
                op_tags=self.op_tags,
                automation_condition=self.automation_condition,
                _required_resource_keys=self._required_resource_keys,
            )

    def get_resource_requirements(self) -> Iterator[ResourceRequirement]:
        if self.node_def is not None:
            yield from self.node_def.get_resource_requirements(
                asset_layer=None,
                handle=None,
            )
        yield SourceAssetIOManagerRequirement(
            key=self.get_io_manager_key(), asset_key=self.key.to_string()
        )
        for source_key, resource_def in self.resource_defs.items():
            yield from resource_def.get_resource_requirements(source_key=source_key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SourceAsset):
            return False
        else:
            return (
                self.key == other.key
                and self.raw_metadata == other.raw_metadata
                and self.io_manager_key == other.io_manager_key
                and self.description == other.description
                and self.group_name == other.group_name
                and self.resource_defs == other.resource_defs
                and self.observe_fn == other.observe_fn
            )



from abc import ABC
from collections.abc import Sequence
from inspect import Parameter
from typing import Annotated, Any, Optional, TypeVar

from dagster._core.decorator_utils import get_function_params, get_type_hints
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._seven import is_subclass


def get_resource_args(fn) -> Sequence[Parameter]:
    type_annotations = get_type_hints(fn)
    return [
        param
        for param in get_function_params(fn)
        if _is_resource_annotation(type_annotations.get(param.name))
    ]


RESOURCE_PARAM_METADATA = "resource_param"


class TreatAsResourceParam(ABC):
    """Marker class for types that can be used as a parameter on an annotated
    function like `@asset`. Any type marked with this class does not require
    a ResourceParam when used on an asset.

    Example:
        class YourClass(TreatAsResourceParam):
            ...

        @asset
        def an_asset(your_class: YourClass):
            ...
    """


def _is_resource_annotation(annotation: Optional[type[Any]]) -> bool:
    from dagster._config.pythonic_config import ConfigurableResourceFactory

    if isinstance(annotation, type) and (
        is_subclass(annotation, ResourceDefinition)
        or is_subclass(annotation, ConfigurableResourceFactory)
        or is_subclass(annotation, TreatAsResourceParam)
    ):
        return True

    return hasattr(annotation, "__metadata__") and getattr(annotation, "__metadata__") == (
        RESOURCE_PARAM_METADATA,
    )


T = TypeVar("T")
ResourceParam = Annotated[T, RESOURCE_PARAM_METADATA]



from dagster._core.run_coordinator.sync_in_memory_run_coordinator import SyncInMemoryRunCoordinator

# for backwards compatibility, we need to keep the old DefaultRunCoordinator
DefaultRunCoordinator = SyncInMemoryRunCoordinator



from dagster._core.storage.event_log.base import (
    AssetRecord as AssetRecord,
    EventLogRecord as EventLogRecord,
    EventLogStorage as EventLogStorage,
)
from dagster._core.storage.event_log.in_memory import (
    InMemoryEventLogStorage as InMemoryEventLogStorage,
)
from dagster._core.storage.event_log.polling_event_watcher import (
    SqlPollingEventWatcher as SqlPollingEventWatcher,
)
from dagster._core.storage.event_log.schema import (
    AssetKeyTable as AssetKeyTable,
    DynamicPartitionsTable as DynamicPartitionsTable,
    SqlEventLogStorageMetadata as SqlEventLogStorageMetadata,
    SqlEventLogStorageTable as SqlEventLogStorageTable,
)
from dagster._core.storage.event_log.sql_event_log import SqlEventLogStorage as SqlEventLogStorage
from dagster._core.storage.event_log.sqlite import (
    ConsolidatedSqliteEventLogStorage as ConsolidatedSqliteEventLogStorage,
    SqliteEventLogStorage as SqliteEventLogStorage,
)



import logging
import sys
import time
from collections.abc import Sequence
from typing import AbstractSet, NamedTuple, Optional  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.errors import DagsterInvalidSubsetError, DagsterUserCodeProcessError
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import RemoteExecutionPlan, RemoteJob
from dagster._core.snap import ExecutionPlanSnapshot
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.workspace.context import BaseWorkspaceRequestContext, IWorkspaceProcessContext
from dagster._utils import SingleInstigatorDebugCrashFlags, check_for_debug_crash

EXECUTION_PLAN_CREATION_RETRIES = 1


class RunRequestExecutionData(NamedTuple):
    remote_job: RemoteJob
    remote_execution_plan: RemoteExecutionPlan


def _get_implicit_job_name_for_assets(
    asset_graph: RemoteWorkspaceAssetGraph, asset_keys: Sequence[AssetKey]
) -> Optional[str]:
    job_names = set(asset_graph.get_materialization_job_names(asset_keys[0]))
    for asset_key in asset_keys[1:]:
        job_names &= set(asset_graph.get_materialization_job_names(asset_key))

    return next(
        (job_name for job_name in job_names if job_name.startswith(IMPLICIT_ASSET_JOB_NAME)), None
    )


def _get_execution_plan_entity_keys(
    execution_plan_snapshot: ExecutionPlanSnapshot,
) -> AbstractSet[EntityKey]:
    output_entity_keys = set()
    for step in execution_plan_snapshot.steps:
        if step.key in execution_plan_snapshot.step_keys_to_execute:
            for output in step.outputs:
                output_properties = check.not_none(output.properties)
                if output_properties.asset_key:
                    output_entity_keys.add(output_properties.asset_key)
                if output_properties.asset_check_key:
                    output_entity_keys.add(output_properties.asset_check_key)
    return output_entity_keys


def _get_job_execution_data_from_run_request(
    asset_graph: RemoteWorkspaceAssetGraph,
    run_request: RunRequest,
    instance: DagsterInstance,
    workspace: BaseWorkspaceRequestContext,
    run_request_execution_data_cache: dict[JobSubsetSelector, RunRequestExecutionData],
) -> RunRequestExecutionData:
    check.invariant(
        len(run_request.entity_keys) > 0,
        "Expected RunRequest to have an asset selection or asset check keys",
    )
    handle = asset_graph.get_repository_handle(run_request.entity_keys[0])
    location_name = handle.location_name
    job_name = (
        _get_implicit_job_name_for_assets(asset_graph, run_request.asset_selection)
        if run_request.asset_selection
        # if we're only executing checks, then this must have been created after the single implicit
        # asset job changes, so we don't need to do the more exhaustive check
        else IMPLICIT_ASSET_JOB_NAME
    )
    if job_name is None:
        check.failed(
            "Could not find an implicit asset job for the given assets:"
            f" {run_request.asset_selection}"
        )

    pipeline_selector = JobSubsetSelector(
        location_name=location_name,
        repository_name=handle.repository_name,
        job_name=job_name,
        asset_selection=run_request.asset_selection,
        asset_check_selection=run_request.asset_check_keys,
        op_selection=None,
    )

    if pipeline_selector not in run_request_execution_data_cache:
        code_location = workspace.get_code_location(handle.location_name)
        remote_job = code_location.get_job(pipeline_selector)

        remote_execution_plan = code_location.get_execution_plan(
            remote_job,
            {},
            step_keys_to_execute=None,
            known_state=None,
            instance=instance,
        )

        run_request_execution_data_cache[pipeline_selector] = RunRequestExecutionData(
            remote_job,
            remote_execution_plan,
        )

    return run_request_execution_data_cache[pipeline_selector]


def _create_asset_run(
    run_id: Optional[str],
    run_request: RunRequest,
    run_request_index: int,
    instance: DagsterInstance,
    run_request_execution_data_cache: dict[JobSubsetSelector, RunRequestExecutionData],
    workspace_process_context: IWorkspaceProcessContext,
    debug_crash_flags: SingleInstigatorDebugCrashFlags,
    logger: logging.Logger,
) -> DagsterRun:
    """Creates a run on the instance for the given run request. Ensures that the created run results
    in an ExecutionPlan that targets the given asset selection. If it does not, attempts to create
    a valid ExecutionPlan by reloading the workspace.
    """
    from dagster._daemon.controller import RELOAD_WORKSPACE_INTERVAL

    if not run_request.asset_selection and not run_request.asset_check_keys:
        check.failed("Expected RunRequest to have an asset selection")

    for i in range(EXECUTION_PLAN_CREATION_RETRIES + 1):
        should_retry = False
        execution_data = None

        # retry until the execution plan targets the asset selection
        try:
            # create a new request context for each run in case the code location server
            # is swapped out in the middle of the submission process
            workspace = workspace_process_context.create_request_context()
            asset_graph = workspace.asset_graph
            execution_data = _get_job_execution_data_from_run_request(
                asset_graph,
                run_request,
                instance,
                workspace=workspace,
                run_request_execution_data_cache=run_request_execution_data_cache,
            )

        except (DagsterInvalidSubsetError, DagsterUserCodeProcessError) as e:
            # Only retry on DagsterInvalidSubsetErrors raised within the code server
            if isinstance(e, DagsterUserCodeProcessError) and not any(
                error_info.cls_name == "DagsterInvalidSubsetError"
                for error_info in e.user_code_process_error_infos
            ):
                raise

            logger.warning(
                "Error while generating the execution plan, possibly because the code server is "
                "out of sync with the daemon. The daemon periodically refreshes its representation "
                f"of the workspace every {RELOAD_WORKSPACE_INTERVAL} seconds - pausing long enough "
                "to ensure that that refresh will happen to bring them back in sync.",
                exc_info=sys.exc_info(),
            )
            should_retry = True

        check_for_debug_crash(debug_crash_flags, "EXECUTION_PLAN_CREATED")
        check_for_debug_crash(debug_crash_flags, f"EXECUTION_PLAN_CREATED_{run_request_index}")

        if not should_retry:
            execution_plan_entity_keys = _get_execution_plan_entity_keys(
                check.not_none(execution_data).remote_execution_plan.execution_plan_snapshot
            )

            if not all(
                key in execution_plan_entity_keys
                for key in [
                    *(run_request.asset_selection or []),
                    *(run_request.asset_check_keys or []),
                ]
            ):
                logger.warning(
                    f"Execution plan targeted the following keys: {execution_plan_entity_keys}, "
                    "which did not include all assets / checks on the run request, "
                    "possibly because the code server is out of sync with the daemon. The daemon "
                    "periodically refreshes its representation of the workspace every "
                    f"{RELOAD_WORKSPACE_INTERVAL} seconds - pausing long enough "
                    "to ensure that that refresh will happen to bring them back in sync.",
                )

                should_retry = True

        if not should_retry:
            remote_job = check.not_none(execution_data).remote_job
            remote_execution_plan = check.not_none(execution_data).remote_execution_plan

            run = instance.create_run(
                job_snapshot=remote_job.job_snapshot,
                execution_plan_snapshot=remote_execution_plan.execution_plan_snapshot,
                parent_job_snapshot=remote_job.parent_job_snapshot,
                job_name=remote_job.name,
                run_id=run_id,
                resolved_op_selection=None,
                op_selection=None,
                run_config={},
                step_keys_to_execute=None,
                tags=run_request.tags,
                root_run_id=None,
                parent_run_id=None,
                status=DagsterRunStatus.NOT_STARTED,
                remote_job_origin=remote_job.get_remote_origin(),
                job_code_origin=remote_job.get_python_origin(),
                asset_selection=frozenset(run_request.asset_selection)
                if run_request.asset_selection
                else None,
                asset_check_selection=frozenset(run_request.asset_check_keys)
                if run_request.asset_check_keys
                else None,
                asset_graph=asset_graph,
            )

            return run

        if i < EXECUTION_PLAN_CREATION_RETRIES:
            # Sleep for RELOAD_WORKSPACE_INTERVAL seconds since the workspace can be refreshed
            # at most once every interval
            time.sleep(RELOAD_WORKSPACE_INTERVAL)
            # Clear the execution plan cache as this data is no longer valid
            run_request_execution_data_cache = {}

            # If the execution plan does not targets the asset selection, the asset graph
            # likely is outdated and targeting the wrong job, refetch the asset
            # graph from the workspace
            workspace = workspace_process_context.create_request_context()
            asset_graph = workspace.asset_graph

    check.failed(
        f"Failed to target asset selection {run_request.asset_selection} in run after retrying."
    )


def submit_asset_run(
    run_id: Optional[str],
    run_request: RunRequest,
    run_request_index: int,
    instance: DagsterInstance,
    workspace_process_context: IWorkspaceProcessContext,
    run_request_execution_data_cache: dict[JobSubsetSelector, RunRequestExecutionData],
    debug_crash_flags: SingleInstigatorDebugCrashFlags,
    logger: logging.Logger,
) -> DagsterRun:
    """Submits a run for a run request that targets an asset selection. If the run already exists,
    submits the existing run. If the run does not exist, creates a new run and submits it, ensuring
    that the created run targets the given asset selection.
    """
    check.invariant(not run_request.run_config, "Asset run requests have no custom run config")
    entity_keys: Sequence[EntityKey] = [
        *(run_request.asset_selection or []),
        *(run_request.asset_check_keys or []),
    ]

    check.invariant(len(entity_keys) > 0)

    # check if the run already exists
    existing_run = instance.get_run_by_id(run_id) if run_id else None
    if existing_run:
        if existing_run.status != DagsterRunStatus.NOT_STARTED:
            logger.warn(
                f"Run {run_id} already submitted on a previously interrupted tick, skipping"
            )

            check_for_debug_crash(debug_crash_flags, "RUN_SUBMITTED")
            check_for_debug_crash(debug_crash_flags, f"RUN_SUBMITTED_{run_request_index}")

            return existing_run
        else:
            logger.warn(
                f"Run {run_id} already created on a previously interrupted tick, submitting"
            )
            run_to_submit = existing_run
    else:
        run_to_submit = _create_asset_run(
            run_id,
            run_request,
            run_request_index,
            instance,
            run_request_execution_data_cache,
            workspace_process_context,
            debug_crash_flags,
            logger,
        )

    check_for_debug_crash(debug_crash_flags, "RUN_CREATED")
    check_for_debug_crash(debug_crash_flags, f"RUN_CREATED_{run_request_index}")

    instance.submit_run(run_to_submit.run_id, workspace_process_context.create_request_context())

    check_for_debug_crash(debug_crash_flags, "RUN_SUBMITTED")
    check_for_debug_crash(debug_crash_flags, f"RUN_SUBMITTED_{run_request_index}")

    asset_key_str = ", ".join([key.to_user_string() for key in entity_keys])

    logger.info(
        f"Submitted run {run_to_submit.run_id} for assets/checks {asset_key_str} with tags"
        f" {run_request.tags}"
    )

    return run_to_submit



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Optional

import sqlalchemy as db
from packaging.version import parse
from sqlalchemy.engine import Connection
from sqlalchemy.pool import NullPool

from dagster import (
    StringSource,
    _check as check,
)
from dagster._config.config_schema import UserConfigSchema
from dagster._core.storage.schedules.schema import ScheduleStorageSqlMetadata
from dagster._core.storage.schedules.sql_schedule_storage import SqlScheduleStorage
from dagster._core.storage.sql import (
    AlembicVersion,
    check_alembic_revision,
    create_engine,
    get_alembic_config,
    run_alembic_upgrade,
    safe_commit,
    stamp_alembic_rev,
)
from dagster._core.storage.sqlite import (
    LAST_KNOWN_STAMPED_SQLITE_ALEMBIC_REVISION,
    create_db_conn_string,
    get_sqlite_version,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster._utils import mkdir_p

MINIMUM_SQLITE_BATCH_VERSION = "3.25.0"


class SqliteScheduleStorage(SqlScheduleStorage, ConfigurableClass):
    """Local SQLite backed schedule storage."""

    def __init__(self, conn_string: str, inst_data: Optional[ConfigurableClassData] = None):
        check.str_param(conn_string, "conn_string")
        self._conn_string = conn_string
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)

        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {"base_dir": StringSource}

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value
    ) -> "SqliteScheduleStorage":
        return SqliteScheduleStorage.from_local(inst_data=inst_data, **config_value)

    @classmethod
    def from_local(
        cls, base_dir: str, inst_data: Optional[ConfigurableClassData] = None
    ) -> "SqliteScheduleStorage":
        check.str_param(base_dir, "base_dir")
        mkdir_p(base_dir)
        conn_string = create_db_conn_string(base_dir, "schedules")
        engine = create_engine(conn_string, poolclass=NullPool)
        alembic_config = get_alembic_config(__file__)

        should_migrate_data = False
        with engine.connect() as connection:
            db_revision, head_revision = check_alembic_revision(alembic_config, connection)
            if not (db_revision and head_revision):
                table_names = db.inspect(engine).get_table_names()
                if "job_ticks" in table_names:
                    # The ticks table exists but the alembic version table does not. This means that the SQLite db was
                    # initialized with SQLAlchemy 2.0 before https://github.com/dagster-io/dagster/pull/25740 was merged.
                    # We should pin the alembic revision to the last known stamped revision before we unpinned SQLAlchemy 2.0
                    # This should be safe because we have guarded all known migrations since then.
                    rev_to_stamp = LAST_KNOWN_STAMPED_SQLITE_ALEMBIC_REVISION
                else:
                    should_migrate_data = True
                    rev_to_stamp = "head"
                ScheduleStorageSqlMetadata.create_all(engine)
                connection.execute(db.text("PRAGMA journal_mode=WAL;"))
                stamp_alembic_rev(alembic_config, connection, rev=rev_to_stamp)
                safe_commit(connection)

        schedule_storage = cls(conn_string, inst_data)
        if should_migrate_data:
            schedule_storage.migrate()
            schedule_storage.optimize()

        return schedule_storage

    @contextmanager
    def connect(self) -> Iterator[Connection]:
        engine = create_engine(self._conn_string, poolclass=NullPool)
        with engine.connect() as conn:
            with conn.begin():
                yield conn

    @property
    def supports_batch_queries(self) -> bool:
        if not super().supports_batch_queries:
            return False

        return super().supports_batch_queries and parse(get_sqlite_version()) >= parse(
            MINIMUM_SQLITE_BATCH_VERSION
        )

    def upgrade(self) -> None:
        alembic_config = get_alembic_config(__file__)
        with self.connect() as conn:
            run_alembic_upgrade(alembic_config, conn)

    def alembic_version(self) -> AlembicVersion:
        alembic_config = get_alembic_config(__file__)
        with self.connect() as conn:
            return check_alembic_revision(alembic_config, conn)



import inspect
from collections.abc import Iterable, Sequence
from importlib import import_module
from types import ModuleType
from typing import Optional

import dagster._check as check
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.asset_key import (
    CoercibleToAssetKeyPrefix,
    check_opt_coercible_to_asset_key_prefix_param,
)
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.module_loaders.object_list import ModuleScopedDagsterDefs
from dagster._core.definitions.module_loaders.utils import find_modules_in_package


def load_asset_checks_from_modules(
    modules: Iterable[ModuleType],
    asset_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
) -> Sequence[AssetChecksDefinition]:
    """Constructs a list of asset checks from the given modules. This is most often used in
    conjunction with a call to `load_assets_from_modules`.

    Args:
        modules (Iterable[ModuleType]): The Python modules to look for checks inside.
        asset_key_prefix (Optional[Union[str, Sequence[str]]]):
            The prefix for the asset keys targeted by the loaded checks. This should match the
            key_prefix argument to load_assets_from_modules.

    Returns:
        Sequence[AssetChecksDefinition]:
            A list containing asset checks defined in the given modules.
    """
    asset_key_prefix = check_opt_coercible_to_asset_key_prefix_param(
        asset_key_prefix, "asset_key_prefix"
    )
    return (
        ModuleScopedDagsterDefs.from_modules(modules, types_to_load=(AssetsDefinition,))
        .get_object_list()
        .with_attributes(
            key_prefix=asset_key_prefix,
            source_key_prefix=None,
            group_name=None,
            freshness_policy=None,
            automation_condition=None,
            backfill_policy=None,
        )
        .checks_defs
    )


def load_asset_checks_from_current_module(
    asset_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
) -> Sequence[AssetChecksDefinition]:
    """Constructs a list of asset checks from the module where this function is called. This is most
    often used in conjunction with a call to `load_assets_from_current_module`.

    Args:
        asset_key_prefix (Optional[Union[str, Sequence[str]]]):
            The prefix for the asset keys targeted by the loaded checks. This should match the
            key_prefix argument to load_assets_from_current_module.

    Returns:
        Sequence[AssetChecksDefinition]:
            A list containing asset checks defined in the current module.
    """
    caller = inspect.stack()[1]
    module = inspect.getmodule(caller[0])
    if module is None:
        check.failed("Could not find a module for the caller")

    asset_key_prefix = check_opt_coercible_to_asset_key_prefix_param(
        asset_key_prefix, "asset_key_prefix"
    )

    return load_asset_checks_from_modules([module], asset_key_prefix=asset_key_prefix)


def load_asset_checks_from_package_module(
    package_module: ModuleType, asset_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None
) -> Sequence[AssetChecksDefinition]:
    """Constructs a list of asset checks from all sub-modules of the given package module. This is
    most often used in conjunction with a call to `load_assets_from_package_module`.

    Args:
        package_module (ModuleType): The Python module to look for checks inside.
        asset_key_prefix (Optional[Union[str, Sequence[str]]]):
            The prefix for the asset keys targeted by the loaded checks. This should match the
            key_prefix argument to load_assets_from_package_module.

    Returns:
        Sequence[AssetChecksDefinition]:
            A list containing asset checks defined in the package.
    """
    asset_key_prefix = check_opt_coercible_to_asset_key_prefix_param(
        asset_key_prefix, "asset_key_prefix"
    )

    return load_asset_checks_from_modules(
        find_modules_in_package(package_module), asset_key_prefix=asset_key_prefix
    )


def load_asset_checks_from_package_name(
    package_name: str, asset_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None
) -> Sequence[AssetChecksDefinition]:
    """Constructs a list of asset checks from all sub-modules of the given package. This is most
    often used in conjunction with a call to `load_assets_from_package_name`.

    Args:
        package_name (str): The name of the Python package to look for checks inside.
        asset_key_prefix (Optional[Union[str, Sequence[str]]]):
            The prefix for the asset keys targeted by the loaded checks. This should match the
            key_prefix argument to load_assets_from_package_name.

    Returns:
        Sequence[AssetChecksDefinition]:
            A list containing asset checks defined in the package.
    """
    asset_key_prefix = check_opt_coercible_to_asset_key_prefix_param(
        asset_key_prefix, "asset_key_prefix"
    )

    package_module = import_module(package_name)
    return load_asset_checks_from_modules(
        find_modules_in_package(package_module), asset_key_prefix=asset_key_prefix
    )



from dagster._core.scheduler.execution import (
    ScheduledExecutionFailed as ScheduledExecutionFailed,
    ScheduledExecutionResult as ScheduledExecutionResult,
    ScheduledExecutionSkipped as ScheduledExecutionSkipped,
    ScheduledExecutionSuccess as ScheduledExecutionSuccess,
)
from dagster._core.scheduler.scheduler import (
    DagsterDaemonScheduler as DagsterDaemonScheduler,
    DagsterScheduleDoesNotExist as DagsterScheduleDoesNotExist,
    DagsterSchedulerError as DagsterSchedulerError,
    Scheduler as Scheduler,
    SchedulerDebugInfo as SchedulerDebugInfo,
)



from typing import TYPE_CHECKING, Optional

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.security import non_secure_md5_hash_str

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_automation.automation_context import (
        AutomationContext,
    )


@whitelist_for_serdes
@record
class RuleCondition(BuiltinAutomationCondition[AssetKey]):
    """This class represents the condition that a particular AutoMaterializeRule is satisfied."""

    rule: AutoMaterializeRule

    def get_node_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[str]) -> str:
        # preserves old (bad) behavior of not including the parent_unique_id to avoid inavlidating
        # old serialized information
        parts = [self.rule.__class__.__name__, self.description]
        return non_secure_md5_hash_str("".join(parts).encode())

    @property
    def description(self) -> str:
        return self.rule.description

    def evaluate(self, context: "AutomationContext[AssetKey]") -> AutomationResult[AssetKey]:
        context.log.debug(f"Evaluating rule: {self.rule.to_snapshot()}")
        # Allow for access to legacy context in legacy rule evaluation
        evaluation_result = self.rule.evaluate_for_asset(context)
        context.log.debug(
            f"Rule returned {evaluation_result.true_subset.size} partitions "
            f"({evaluation_result.end_timestamp - evaluation_result.start_timestamp:.2f} seconds)"
        )
        return evaluation_result



from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Any, NamedTuple, Optional  # noqa: UP035

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_graph import AssetGraph
from dagster._core.definitions.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.cacheable_assets import AssetsDefinitionCacheableData
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.metadata import MetadataMapping
from dagster._core.definitions.metadata.metadata_value import (
    CodeLocationReconstructionMetadataValue,
)
from dagster._core.definitions.repository_definition.repository_data import RepositoryData
from dagster._core.definitions.repository_definition.valid_definitions import (
    RepositoryElementDefinition as RepositoryElementDefinition,
)
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import check_valid_name
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.instance import DagsterInstance
from dagster._serdes import whitelist_for_serdes
from dagster._utils import hash_collection
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.definitions import AssetsDefinition
    from dagster._core.definitions.asset_checks import AssetChecksDefinition
    from dagster._core.storage.asset_value_loader import AssetValueLoader


@whitelist_for_serdes(
    storage_field_names={"cacheable_asset_data": "cached_data_by_key"},
)
class RepositoryLoadData(
    NamedTuple(
        "_RepositoryLoadData",
        [
            ("cacheable_asset_data", Mapping[str, Sequence[AssetsDefinitionCacheableData]]),
            ("reconstruction_metadata", Mapping[str, Any]),
        ],
    )
):
    def __new__(
        cls,
        cacheable_asset_data: Optional[
            Mapping[str, Sequence[AssetsDefinitionCacheableData]]
        ] = None,
        reconstruction_metadata: Optional[
            Mapping[str, CodeLocationReconstructionMetadataValue]
        ] = None,
    ):
        return super().__new__(
            cls,
            cacheable_asset_data=(
                check.opt_mapping_param(
                    cacheable_asset_data,
                    "cacheable_asset_data",
                    key_type=str,
                    value_type=list,
                )
            ),
            reconstruction_metadata=check.opt_mapping_param(
                reconstruction_metadata, "reconstruction_metadata", key_type=str
            ),
        )

    # Allow this to be hashed for use in `lru_cache`. This is needed because:
    # - `ReconstructableJob` uses `lru_cache`
    # - `ReconstructableJob` has a `ReconstructableRepository` attribute
    # - `ReconstructableRepository` has a `RepositoryLoadData` attribute
    # - `RepositoryLoadData` has collection attributes that are unhashable by default
    def __hash__(self) -> int:
        if not hasattr(self, "_hash"):
            self._hash = hash_collection(self)
        return self._hash

    def replace_reconstruction_metadata(
        self, reconstruction_metadata: Mapping[str, Any]
    ) -> "RepositoryLoadData":
        return RepositoryLoadData(
            cacheable_asset_data=self.cacheable_asset_data,
            reconstruction_metadata=reconstruction_metadata,
        )


class RepositoryDefinition:
    """Define a repository that contains a group of definitions.

    Users should typically not create objects of this class directly. Instead, use the
    :py:func:`@repository` decorator.

    Args:
        name (str): The name of the repository.
        repository_data (RepositoryData): Contains the definitions making up the repository.
        description (Optional[str]): A string description of the repository.
        metadata (Optional[MetadataMapping]): Arbitrary metadata for the repository. Not
            displayed in the UI but accessible on RepositoryDefinition at runtime.
    """

    def __init__(
        self,
        name: str,
        *,
        repository_data: RepositoryData,
        description: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        repository_load_data: Optional[RepositoryLoadData] = None,
    ):
        self._name = check_valid_name(name)
        self._description = check.opt_str_param(description, "description")
        self._repository_data = check.inst_param(repository_data, "repository_data", RepositoryData)
        self._metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
        self._repository_load_data = check.opt_inst_param(
            repository_load_data, "repository_load_data", RepositoryLoadData
        )

    @property
    def repository_load_data(self) -> Optional[RepositoryLoadData]:
        return self._repository_load_data

    def replace_reconstruction_metadata(
        self, reconstruction_metadata: Mapping[str, str]
    ) -> "RepositoryDefinition":
        """Modifies the repository load data to include the provided reconstruction metadata."""
        check.mapping_param(reconstruction_metadata, "reconstruction_metadata", key_type=str)
        if not reconstruction_metadata:
            return self

        for k, v in reconstruction_metadata.items():
            if not isinstance(v, str):
                raise DagsterInvariantViolationError(
                    f"Reconstruction metadata values must be strings. State-representing values are"
                    f" expected to be serialized before being passed as reconstruction metadata."
                    f" Got for key {k}:\n\n{v}"
                )
        normalized_metadata = {
            k: CodeLocationReconstructionMetadataValue(v)
            for k, v in reconstruction_metadata.items()
        }
        return RepositoryDefinition(
            self._name,
            repository_data=self._repository_data,
            description=self._description,
            metadata=self._metadata,
            repository_load_data=self._repository_load_data.replace_reconstruction_metadata(
                normalized_metadata
            )
            if self._repository_load_data
            else RepositoryLoadData(reconstruction_metadata=normalized_metadata),
        )

    @public
    @property
    def name(self) -> str:
        """str: The name of the repository."""
        return self._name

    @public
    @property
    def description(self) -> Optional[str]:
        """Optional[str]: A human-readable description of the repository."""
        return self._description

    @public
    @property
    def metadata(self) -> Optional[MetadataMapping]:
        """Optional[MetadataMapping]: Arbitrary metadata for the repository."""
        return self._metadata

    def load_all_definitions(self) -> None:
        # force load of all lazy constructed code artifacts
        self._repository_data.load_all_definitions()

    @public
    @property
    def job_names(self) -> Sequence[str]:
        """List[str]: Names of all jobs in the repository."""
        return self._repository_data.get_job_names()

    def get_top_level_resources(self) -> Mapping[str, ResourceDefinition]:
        return self._repository_data.get_top_level_resources()

    def get_env_vars_by_top_level_resource(self) -> Mapping[str, AbstractSet[str]]:
        return self._repository_data.get_env_vars_by_top_level_resource()

    @public
    def has_job(self, name: str) -> bool:
        """Check if a job with a given name is present in the repository.

        Args:
            name (str): The name of the job.

        Returns:
            bool
        """
        return self._repository_data.has_job(name)

    @public
    def get_job(self, name: str) -> JobDefinition:
        """Get a job by name.

        If this job is present in the lazily evaluated dictionary passed to the
        constructor, but has not yet been constructed, only this job is constructed, and
        will be cached for future calls.

        Args:
            name (str): Name of the job to retrieve.

        Returns:
            JobDefinition: The job definition corresponding to
            the given name.
        """
        return self._repository_data.get_job(name)

    @public
    def get_all_jobs(self) -> Sequence[JobDefinition]:
        """Return all jobs in the repository as a list.

        Note that this will construct any job in the lazily evaluated dictionary that has
        not yet been constructed.

        Returns:
            List[JobDefinition]: All jobs in the repository.
        """
        return self._repository_data.get_all_jobs()

    @public
    @property
    def schedule_defs(self) -> Sequence[ScheduleDefinition]:
        """List[ScheduleDefinition]: All schedules in the repository."""
        return self._repository_data.get_all_schedules()

    @public
    def get_schedule_def(self, name: str) -> ScheduleDefinition:
        """Get a schedule definition by name.

        Args:
            name (str): The name of the schedule.

        Returns:
            ScheduleDefinition: The schedule definition.
        """
        return self._repository_data.get_schedule(name)

    @public
    def has_schedule_def(self, name: str) -> bool:
        """bool: Check if a schedule with a given name is present in the repository."""
        return self._repository_data.has_schedule(name)

    @public
    @property
    def sensor_defs(self) -> Sequence[SensorDefinition]:
        """Sequence[SensorDefinition]: All sensors in the repository."""
        return self._repository_data.get_all_sensors()

    @public
    def get_sensor_def(self, name: str) -> SensorDefinition:
        """Get a sensor definition by name.

        Args:
            name (str): The name of the sensor.

        Returns:
            SensorDefinition: The sensor definition.
        """
        return self._repository_data.get_sensor(name)

    @public
    def has_sensor_def(self, name: str) -> bool:
        """bool: Check if a sensor with a given name is present in the repository."""
        return self._repository_data.has_sensor(name)

    @public
    @property
    def source_assets_by_key(self) -> Mapping[AssetKey, SourceAsset]:
        """Mapping[AssetKey, SourceAsset]: The source assets defined in the repository."""
        return self._repository_data.get_source_assets_by_key()

    # NOTE: `assets_defs_by_key` should generally not be used internally. It returns the
    # `AssetsDefinition` supplied at repository construction time. Internally, assets defs should be
    # obtained from the `AssetGraph` via `asset_graph.assets_defs`. This returns a normalized set of
    # assets defs, where:
    #
    # - `SourceAsset` instances are replaced with external `AssetsDefinition`
    # - Relative asset key dependencies are resolved
    # - External `AssetsDefinition` have been generated for referenced asset keys without a
    #   corresponding user-provided definition

    @public
    @property
    def assets_defs_by_key(self) -> Mapping[AssetKey, "AssetsDefinition"]:
        """Mapping[AssetKey, AssetsDefinition]: The assets definitions defined in the repository."""
        return self._repository_data.get_assets_defs_by_key()

    @public
    @property
    def asset_checks_defs_by_key(self) -> Mapping[AssetCheckKey, "AssetChecksDefinition"]:
        """Mapping[AssetCheckKey, AssetChecksDefinition]: The assets checks defined in the repository."""
        return self._repository_data.get_asset_checks_defs_by_key()

    def has_implicit_global_asset_job_def(self) -> bool:
        return self.has_job(IMPLICIT_ASSET_JOB_NAME)

    def get_implicit_global_asset_job_def(self) -> JobDefinition:
        return self.get_job(IMPLICIT_ASSET_JOB_NAME)

    def get_implicit_asset_job_names(self) -> Sequence[str]:
        return [IMPLICIT_ASSET_JOB_NAME]

    def get_implicit_job_def_for_assets(
        self, asset_keys: Iterable[AssetKey]
    ) -> Optional[JobDefinition]:
        return self.get_job(IMPLICIT_ASSET_JOB_NAME)

    def get_maybe_subset_job_def(
        self,
        job_name: str,
        op_selection: Optional[Iterable[str]] = None,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
        asset_check_selection: Optional[AbstractSet[AssetCheckKey]] = None,
    ):
        defn = self.get_job(job_name)
        return defn.get_subset(
            op_selection=op_selection,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
        )

    @public
    def load_asset_value(
        self,
        asset_key: CoercibleToAssetKey,
        *,
        python_type: Optional[type] = None,
        instance: Optional[DagsterInstance] = None,
        partition_key: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        resource_config: Optional[Any] = None,
    ) -> object:
        """Load the contents of an asset as a Python object.

        Invokes `load_input` on the :py:class:`IOManager` associated with the asset.

        If you want to load the values of multiple assets, it's more efficient to use
        :py:meth:`~dagster.RepositoryDefinition.get_asset_value_loader`, which avoids spinning up
        resources separately for each asset.

        Args:
            asset_key (Union[AssetKey, Sequence[str], str]): The key of the asset to load.
            python_type (Optional[Type]): The python type to load the asset as. This is what will
                be returned inside `load_input` by `context.dagster_type.typing_type`.
            partition_key (Optional[str]): The partition of the asset to load.
            metadata (Optional[Dict[str, Any]]): Input metadata to pass to the :py:class:`IOManager`
                (is equivalent to setting the metadata argument in `In` or `AssetIn`).
            resource_config (Optional[Any]): A dictionary of resource configurations to be passed
                to the :py:class:`IOManager`.

        Returns:
            The contents of an asset as a Python object.
        """
        from dagster._core.storage.asset_value_loader import AssetValueLoader

        # The normalized assets defs must be obtained from the asset graph, not the repository data
        normalized_assets_defs_by_key = {
            k: ad for ad in self.asset_graph.assets_defs for k in ad.keys
        }
        with AssetValueLoader(normalized_assets_defs_by_key, instance=instance) as loader:
            return loader.load_asset_value(
                asset_key,
                python_type=python_type,
                partition_key=partition_key,
                metadata=metadata,
                resource_config=resource_config,
            )

    @public
    def get_asset_value_loader(
        self, instance: Optional[DagsterInstance] = None
    ) -> "AssetValueLoader":
        """Returns an object that can load the contents of assets as Python objects.

        Invokes `load_input` on the :py:class:`IOManager` associated with the assets. Avoids
        spinning up resources separately for each asset.

        Usage:

        .. code-block:: python

            with my_repo.get_asset_value_loader() as loader:
                asset1 = loader.load_asset_value("asset1")
                asset2 = loader.load_asset_value("asset2")

        """
        from dagster._core.storage.asset_value_loader import AssetValueLoader

        # The normalized assets defs must be obtained from the asset graph, not the repository data
        normalized_assets_defs_by_key = {
            k: ad for ad in self.asset_graph.assets_defs for k in ad.keys
        }
        return AssetValueLoader(normalized_assets_defs_by_key, instance=instance)

    @property
    @cached_method
    def asset_graph(self) -> AssetGraph:
        return AssetGraph.from_assets(
            [
                *list(set(self.assets_defs_by_key.values())),
                *self.source_assets_by_key.values(),
                *list(set(self.asset_checks_defs_by_key.values())),
            ],
        )

    # If definition comes from the @repository decorator, then the __call__ method will be
    # overwritten. Therefore, we want to maintain the call-ability of repository definitions.
    def __call__(self, *args, **kwargs):
        return self



from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.resource_invocation import resolve_bound_config
from dagster._core.execution.context.logger import InitLoggerContext, UnboundInitLoggerContext


def logger_invocation_result(logger_def: LoggerDefinition, init_context: UnboundInitLoggerContext):
    """Using the provided context, call the underlying `logger_fn` and return created logger."""
    logger_config = resolve_bound_config(init_context.logger_config, logger_def)

    bound_context = InitLoggerContext(
        logger_config, logger_def, init_context.job_def, init_context.run_id
    )

    return logger_def.logger_fn(bound_context)



from enum import Enum
from typing import Callable, NamedTuple, Optional

import dagster._check as check


class LocationStateChangeEventType(Enum):
    LOCATION_UPDATED = "LOCATION_UPDATED"
    LOCATION_DISCONNECTED = "LOCATION_DISCONNECTED"
    LOCATION_RECONNECTED = "LOCATION_RECONNECTED"
    LOCATION_ERROR = "LOCATION_ERROR"


class LocationStateChangeEvent(
    NamedTuple(
        "_LocationStateChangeEvent",
        [
            ("event_type", LocationStateChangeEventType),
            ("location_name", str),
            ("message", str),
            ("server_id", Optional[str]),
        ],
    )
):
    def __new__(
        cls,
        event_type: LocationStateChangeEventType,
        location_name: str,
        message: str,
        server_id: Optional[str] = None,
    ):
        return super().__new__(
            cls,
            check.inst_param(event_type, "event_type", LocationStateChangeEventType),
            check.str_param(location_name, "location_name"),
            check.str_param(message, "message"),
            check.opt_str_param(server_id, "server_id"),
        )


class LocationStateSubscriber:
    def __init__(self, callback: Callable[[LocationStateChangeEvent], None]):
        check.callable_param(callback, "callback")
        self._callback = callback

    def handle_event(self, event: LocationStateChangeEvent):
        check.inst_param(event, "event", LocationStateChangeEvent)
        self._callback(event)



from collections import defaultdict
from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, AbstractSet, NamedTuple, Optional  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_key import AssetKey, EntityKey
from dagster._core.definitions.dependency import NodeHandle, NodeInputHandle, NodeOutputHandle
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.errors import DagsterInvariantViolationError

if TYPE_CHECKING:
    from dagster._core.definitions.asset_graph import AssetGraph, AssetNode
    from dagster._core.definitions.assets import AssetsDefinition
    from dagster._core.definitions.partition_mapping import PartitionMapping


class AssetLayer(NamedTuple):
    """Stores all of the asset-related information for a Dagster job. Maps each
    input / output in the underlying graph to the asset it represents (if any), and records the
    dependencies between each asset.

    Args:
        asset_key_by_node_input_handle (Mapping[NodeInputHandle, AssetOutputInfo], optional): A mapping
            from a unique input in the underlying graph to the associated AssetKey that it loads from.
        asset_keys_by_node_output_handle (Mapping[NodeOutputHandle, AssetOutputInfo], optional): A mapping
            from a unique output in the underlying graph to the associated AssetOutputInfo.
    """

    asset_graph: "AssetGraph"
    assets_defs_by_node_handle: Mapping[NodeHandle, "AssetsDefinition"]
    asset_keys_by_node_input_handle: Mapping[NodeInputHandle, AssetKey]
    asset_keys_by_node_output_handle: Mapping[NodeOutputHandle, AssetKey]
    check_key_by_node_output_handle: Mapping[NodeOutputHandle, AssetCheckKey]
    node_output_handles_by_asset_check_key: Mapping[AssetCheckKey, NodeOutputHandle]
    check_names_by_asset_key_by_node_handle: Mapping[
        NodeHandle, Mapping[AssetKey, AbstractSet[str]]
    ]
    outer_node_names_by_asset_key: Mapping[AssetKey, str] = {}

    @staticmethod
    def from_graph_and_assets_node_mapping(
        graph_def: GraphDefinition,
        assets_defs_by_outer_node_handle: Mapping[NodeHandle, "AssetsDefinition"],
        asset_graph: "AssetGraph",
    ) -> "AssetLayer":
        """Generate asset info from a GraphDefinition and a mapping from nodes in that graph to the
        corresponding AssetsDefinition objects.

        Args:
            graph_def (GraphDefinition): The graph for the JobDefinition that we're generating
                this AssetLayer for.
            assets_defs_by_outer_node_handle (Mapping[NodeHandle, AssetsDefinition]): A mapping from
                a NodeHandle pointing to the node in the graph where the AssetsDefinition ended up.
        """
        asset_key_by_input: dict[NodeInputHandle, AssetKey] = {}
        asset_keys_by_node_output_handle: dict[NodeOutputHandle, AssetKey] = {}
        check_key_by_output: dict[NodeOutputHandle, AssetCheckKey] = {}

        node_output_handles_by_asset_check_key: Mapping[AssetCheckKey, NodeOutputHandle] = {}
        check_names_by_asset_key_by_node_handle: dict[NodeHandle, dict[AssetKey, set[str]]] = {}
        assets_defs_by_check_key: dict[AssetCheckKey, AssetsDefinition] = {}
        outer_node_names_by_asset_key: dict[AssetKey, str] = {}
        assets_defs_by_op_handle: dict[NodeHandle, AssetsDefinition] = {}

        for node_handle, assets_def in assets_defs_by_outer_node_handle.items():
            computation = check.not_none(assets_def.computation)
            full_node_def = computation.full_node_def
            for input_name, input_asset_key in assets_def.node_keys_by_input_name.items():
                input_handle = NodeInputHandle(node_handle=node_handle, input_name=input_name)
                asset_key_by_input[input_handle] = input_asset_key
                # resolve graph input to list of op inputs that consume it
                node_input_handles = full_node_def.resolve_input_to_destinations(input_handle)
                for node_input_handle in node_input_handles:
                    asset_key_by_input[node_input_handle] = input_asset_key

            for output_name, asset_key in assets_def.node_keys_by_output_name.items():
                # resolve graph output to the op output it comes from
                inner_output_def, inner_node_handle = full_node_def.resolve_output_to_origin(
                    output_name, handle=node_handle
                )
                node_output_handle = NodeOutputHandle(
                    node_handle=inner_node_handle, output_name=inner_output_def.name
                )

                asset_keys_by_node_output_handle[node_output_handle] = asset_key

                destinations = full_node_def.resolve_output_to_destinations(
                    output_name, node_handle
                )
                for destination in destinations:
                    asset_key_by_input[destination] = asset_key
                    if isinstance(full_node_def, GraphDefinition):
                        for op_input_handle in full_node_def.get_node(
                            destination.node_handle.pop()
                        ).definition.resolve_input_to_destinations(destination):
                            asset_key_by_input[op_input_handle] = asset_key
                outer_node_names_by_asset_key[asset_key] = node_handle.name

            if len(assets_def.check_specs_by_output_name) > 0:
                check_names_by_asset_key_by_node_handle[node_handle] = defaultdict(set)

                for output_name, check_spec in assets_def.check_specs_by_output_name.items():
                    (
                        inner_output_def,
                        inner_node_handle,
                    ) = computation.node_def.resolve_output_to_origin(
                        output_name, handle=node_handle
                    )
                    node_output_handle = NodeOutputHandle(
                        node_handle=inner_node_handle, output_name=inner_output_def.name
                    )
                    node_output_handles_by_asset_check_key[check_spec.key] = node_output_handle
                    check_names_by_asset_key_by_node_handle[node_handle][check_spec.asset_key].add(
                        check_spec.name
                    )
                    check_key_by_output[node_output_handle] = check_spec.key

                assets_defs_by_check_key.update({k: assets_def for k in assets_def.check_keys})

            if computation is not None:
                for op_handle in computation.node_def.get_op_handles(parent=node_handle):
                    assets_defs_by_op_handle[op_handle] = assets_def

        return AssetLayer(
            asset_graph=asset_graph,
            asset_keys_by_node_input_handle=asset_key_by_input,
            asset_keys_by_node_output_handle=asset_keys_by_node_output_handle,
            check_key_by_node_output_handle=check_key_by_output,
            assets_defs_by_node_handle=assets_defs_by_op_handle,
            node_output_handles_by_asset_check_key=node_output_handles_by_asset_check_key,
            check_names_by_asset_key_by_node_handle=check_names_by_asset_key_by_node_handle,
            outer_node_names_by_asset_key=outer_node_names_by_asset_key,
        )

    @property
    def executable_asset_keys(self) -> Iterable[AssetKey]:
        return self.asset_graph.executable_asset_keys

    def get(self, asset_key: AssetKey) -> "AssetNode":
        return self.asset_graph.get(asset_key)

    def has(self, asset_key: AssetKey) -> bool:
        return self.asset_graph.has(asset_key)

    def node_output_handle_for_asset(self, asset_key: AssetKey) -> NodeOutputHandle:
        matching_handles = [
            handle
            for handle, handle_asset_key in self.asset_keys_by_node_output_handle.items()
            if handle_asset_key == asset_key
        ]
        check.invariant(len(matching_handles) == 1)
        return matching_handles[0]

    def assets_def_for_node(self, node_handle: NodeHandle) -> Optional["AssetsDefinition"]:
        return self.assets_defs_by_node_handle.get(node_handle)

    def asset_keys_for_node(self, node_handle: NodeHandle) -> AbstractSet[AssetKey]:
        assets_def = self.assets_def_for_node(node_handle)
        return check.not_none(assets_def).keys

    def asset_key_for_node(self, node_handle: NodeHandle) -> AssetKey:
        asset_keys = self.asset_keys_for_node(node_handle)
        if len(asset_keys) > 1:
            raise DagsterInvariantViolationError(
                "Cannot call `asset_key_for_node` in a multi_asset with more than one asset."
                " Multiple asset keys defined."
            )
        return next(iter(asset_keys))

    def get_output_name_for_asset_check(self, asset_check_key: AssetCheckKey) -> str:
        """Output name in the leaf op."""
        return self.node_output_handles_by_asset_check_key[asset_check_key].output_name

    def asset_key_for_input(self, node_handle: NodeHandle, input_name: str) -> Optional[AssetKey]:
        return self.asset_keys_by_node_input_handle.get(
            NodeInputHandle(node_handle=node_handle, input_name=input_name)
        )

    def input_for_asset_key(self, node_handle: NodeHandle, key: AssetKey) -> Optional[str]:
        return next(
            (
                input_handle.input_name
                for input_handle, k in self.asset_keys_by_node_input_handle.items()
                if k == key
            ),
            None,
        )

    def asset_key_for_output(self, node_handle: NodeHandle, output_name: str) -> Optional[AssetKey]:
        return self.asset_keys_by_node_output_handle.get(
            NodeOutputHandle(node_handle=node_handle, output_name=output_name)
        )

    def asset_check_key_for_output(
        self, node_handle: NodeHandle, output_name: str
    ) -> Optional[AssetCheckKey]:
        return self.check_key_by_node_output_handle.get(
            NodeOutputHandle(node_handle=node_handle, output_name=output_name)
        )

    def partition_mapping_for_node_input(
        self, node_handle: NodeHandle, upstream_asset_key: AssetKey
    ) -> Optional["PartitionMapping"]:
        assets_def = self.assets_defs_by_node_handle.get(node_handle)
        if assets_def is not None:
            return assets_def.get_partition_mapping_for_dep(upstream_asset_key)
        else:
            # Can end up here when a non-asset job has an asset as an input
            return None

    def downstream_dep_assets_and_checks(
        self, node_handle: NodeHandle, output_name: str
    ) -> set[EntityKey]:
        """Given the node handle of an op within a graph-backed asset and an output name,
        returns the asset keys dependent on that output.

        For example, for the following asset:

        @op(out={"out_1": Out(is_required=False), "out_2": Out(is_required=False)})
        def two_outputs_op(context):
            return 1, 1


        @op
        def add_one(x):
            return x + 1


        @graph(out={"asset_one": GraphOut(), "asset_two": GraphOut()})
        def my_graph():
            out_1, out_2 = two_outputs_op()
            return {"asset_one": out_1, "asset_two": add_one(out_2)}

        two_assets = AssetsDefinition.from_graph(my_graph)

        Calling downstream_dep_assets with the node handle of two_outputs_op will return:
        - {AssetKey("asset_one")} if output_name="out_1"
        - {AssetKey("asset_two")} if output_name="out_2"

        Calling downstream_dep_assets with node handle add_one will return:
        - {AssetKey("asset_two")} if output_name="result"
        """
        assets_def = self.assets_defs_by_node_handle[node_handle]
        return {
            key
            for key in check.not_none(assets_def.computation).entity_keys_by_dep_op_output_handle[
                NodeOutputHandle(node_handle=node_handle.pop(), output_name=output_name)
            ]
        }

    def upstream_dep_op_handles(self, asset_key: AssetKey) -> AbstractSet[NodeHandle]:
        computation = check.not_none(self.asset_graph.get(asset_key).assets_def.computation)
        op_handles_in_assets_def = computation.dep_op_handles_by_entity_key[asset_key]
        outer_node_handle = NodeHandle(self.outer_node_names_by_asset_key[asset_key], parent=None)
        return {outer_node_handle.with_child(op_handle) for op_handle in op_handles_in_assets_def}



import logging
import math
import os
import sys
import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, cast

import dagster._check as check
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.event_api import EventLogCursor
from dagster._core.events import DagsterEvent, DagsterEventType, EngineEventData
from dagster._core.execution.context.system import PlanOrchestrationContext
from dagster._core.execution.plan.active import ActiveExecution
from dagster._core.execution.plan.instance_concurrency_context import InstanceConcurrencyContext
from dagster._core.execution.plan.objects import StepFailureData
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.retries import RetryMode
from dagster._core.executor.base import Executor
from dagster._core.executor.step_delegating.step_handler.base import StepHandler, StepHandlerContext
from dagster._core.instance import DagsterInstance
from dagster._grpc.types import ExecuteStepArgs
from dagster._time import get_current_datetime
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info

if TYPE_CHECKING:
    from dagster._core.execution.plan.step import ExecutionStep


def _default_sleep_seconds():
    return float(os.environ.get("DAGSTER_STEP_DELEGATING_EXECUTOR_SLEEP_SECONDS", "1.0"))


class StepDelegatingExecutor(Executor):
    """This executor tails the event log for events from the steps that it spins up. It also
    sometimes creates its own events - when it does, that event is automatically written to the
    event log. But we wait until we later tail it from the event log database before yielding it,
    to avoid yielding the same event multiple times to callsites.
    """

    def __init__(
        self,
        step_handler: StepHandler,
        retries: RetryMode,
        sleep_seconds: Optional[float] = None,
        check_step_health_interval_seconds: Optional[int] = None,
        max_concurrent: Optional[int] = None,
        tag_concurrency_limits: Optional[list[dict[str, Any]]] = None,
        should_verify_step: bool = False,
    ):
        self._step_handler = step_handler
        self._retries = retries

        self._max_concurrent = check.opt_int_param(max_concurrent, "max_concurrent")
        self._tag_concurrency_limits = check.opt_list_param(
            tag_concurrency_limits, "tag_concurrency_limits"
        )

        if self._max_concurrent is not None:
            check.invariant(self._max_concurrent > 0, "max_concurrent must be > 0")

        self._sleep_seconds = cast(
            float,
            check.opt_float_param(sleep_seconds, "sleep_seconds", default=_default_sleep_seconds()),
        )
        self._check_step_health_interval_seconds = cast(
            int,
            check.opt_int_param(
                check_step_health_interval_seconds, "check_step_health_interval_seconds", default=20
            ),
        )
        self._should_verify_step = should_verify_step

        self._event_cursor: Optional[str] = None

        self._pop_events_limit = int(os.getenv("DAGSTER_EXECUTOR_POP_EVENTS_LIMIT", "1000"))

    @property
    def retries(self):
        return self._retries

    def _get_pop_events_offset(self, instance: DagsterInstance):
        if "DAGSTER_EXECUTOR_POP_EVENTS_OFFSET" in os.environ:
            return int(os.environ["DAGSTER_EXECUTOR_POP_EVENTS_OFFSET"])
        return instance.event_log_storage.default_run_scoped_event_tailer_offset()

    def _pop_events(
        self, instance: DagsterInstance, run_id: str, seen_storage_ids: set[int]
    ) -> Sequence[DagsterEvent]:
        conn = instance.get_records_for_run(
            run_id,
            self._event_cursor,
            of_type=set(DagsterEventType),
            limit=self._pop_events_limit,
        )

        dagster_events = [
            record.event_log_entry.dagster_event
            for record in conn.records
            if record.event_log_entry.dagster_event and record.storage_id not in seen_storage_ids
        ]

        returned_storage_ids = {record.storage_id for record in conn.records}

        pop_events_offset = self._get_pop_events_offset(instance)

        if not pop_events_offset:
            self._event_cursor = conn.cursor
        elif (
            len(returned_storage_ids) == self._pop_events_limit
            and returned_storage_ids <= seen_storage_ids
        ):
            # Start the next cursor halfway through the returned list
            desired_next_storage_id = conn.records[
                math.ceil(self._pop_events_limit / 2) - 1
            ].storage_id
            self._event_cursor = EventLogCursor.from_storage_id(desired_next_storage_id).to_string()

            logging.getLogger("dagster").warn(
                f"Event tailer query returned a list of {self._pop_events_limit} storage IDs that had already been returned before. Setting the cursor to {desired_next_storage_id} to ensure it advances."
            )
        else:
            cursor_obj = EventLogCursor.parse(conn.cursor)
            check.invariant(
                cursor_obj.is_id_cursor(),
                "Applying a tailer offset only works with an id-based cursor",
            )
            # Apply offset for next query (while also making sure that the cursor doesn't backtrack)
            current_cursor_storage_id = (
                EventLogCursor.parse(self._event_cursor).storage_id() if self._event_cursor else 0
            )
            new_storage_id = max(
                current_cursor_storage_id, cursor_obj.storage_id() - pop_events_offset
            )
            self._event_cursor = EventLogCursor.from_storage_id(new_storage_id).to_string()

        seen_storage_ids.update(returned_storage_ids)

        return dagster_events

    def _get_step_handler_context(
        self, plan_context, steps, active_execution
    ) -> StepHandlerContext:
        return StepHandlerContext(
            instance=plan_context.plan_data.instance,
            plan_context=plan_context,
            steps=steps,
            execute_step_args=ExecuteStepArgs(
                job_origin=plan_context.reconstructable_job.get_python_origin(),
                run_id=plan_context.dagster_run.run_id,
                step_keys_to_execute=[step.key for step in steps],
                instance_ref=plan_context.plan_data.instance.get_ref(),
                retry_mode=self.retries.for_inner_plan(),
                known_state=active_execution.get_known_state(),
                should_verify_step=self._should_verify_step,
                print_serialized_events=False,
            ),
            dagster_run=plan_context.dagster_run,
        )

    def execute(self, plan_context: PlanOrchestrationContext, execution_plan: ExecutionPlan):
        check.inst_param(plan_context, "plan_context", PlanOrchestrationContext)
        check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
        seen_storage_ids = set()

        DagsterEvent.engine_event(
            plan_context,
            f"Starting execution with step handler {self._step_handler.name}.",
            EngineEventData(),
        )
        with InstanceConcurrencyContext(
            plan_context.instance, plan_context.dagster_run
        ) as instance_concurrency_context:
            with ActiveExecution(
                execution_plan,
                retry_mode=self.retries,
                max_concurrent=self._max_concurrent,
                tag_concurrency_limits=self._tag_concurrency_limits,
                instance_concurrency_context=instance_concurrency_context,
            ) as active_execution:
                running_steps: dict[str, ExecutionStep] = {}

                if plan_context.resume_from_failure:
                    DagsterEvent.engine_event(
                        plan_context,
                        "Resuming execution from failure",
                        EngineEventData(),
                    )

                    prior_events = self._pop_events(
                        plan_context.instance,
                        plan_context.run_id,
                        seen_storage_ids,
                    )
                    for dagster_event in prior_events:
                        yield dagster_event

                    possibly_in_flight_steps = active_execution.rebuild_from_events(prior_events)
                    for step in possibly_in_flight_steps:
                        step_handler_context = self._get_step_handler_context(
                            plan_context, [step], active_execution
                        )

                        DagsterEvent.engine_event(
                            step_handler_context.get_step_context(step.key),
                            f"Checking on status of in-progress step {step.key} from previous run",
                            EngineEventData(),
                        )

                        should_retry_step = False
                        health_check = None

                        try:
                            health_check = self._step_handler.check_step_health(
                                step_handler_context
                            )
                        except Exception:
                            # For now we assume that an exception indicates that the step should be resumed.
                            # This should probably be a separate should_resume_step method on the step handler.
                            DagsterEvent.engine_event(
                                step_handler_context.get_step_context(step.key),
                                f"Including {step.key} in the new run since it raised an error"
                                " when checking whether it was running",
                                EngineEventData(
                                    error=serializable_error_info_from_exc_info(sys.exc_info())
                                ),
                            )
                            should_retry_step = True
                        else:
                            if not health_check.is_healthy:
                                DagsterEvent.engine_event(
                                    step_handler_context.get_step_context(step.key),
                                    f"Including step {step.key} in the new run since it is not"
                                    f" currently running: {health_check.unhealthy_reason}",
                                    EngineEventData(),
                                )
                                should_retry_step = True

                        if should_retry_step:
                            # health check failed, launch the step
                            list(
                                self._step_handler.launch_step(
                                    self._get_step_handler_context(
                                        plan_context, [step], active_execution
                                    )
                                )
                            )

                        running_steps[step.key] = step

                last_check_step_health_time = get_current_datetime()

                try:
                    # Order of events is important here. During an interation, we call handle_event, then get_steps_to_execute,
                    # then is_complete. get_steps_to_execute updates the state of ActiveExecution, and without it
                    # is_complete can return true when we're just between steps.
                    while not active_execution.is_complete:
                        if active_execution.check_for_interrupts():
                            active_execution.mark_interrupted()
                            if not plan_context.instance.run_will_resume(plan_context.run_id):
                                DagsterEvent.engine_event(
                                    plan_context,
                                    "Executor received termination signal, forwarding to steps",
                                    EngineEventData.interrupted(list(running_steps.keys())),
                                )
                                for step in running_steps.values():
                                    list(
                                        self._step_handler.terminate_step(
                                            self._get_step_handler_context(
                                                plan_context, [step], active_execution
                                            )
                                        )
                                    )
                            else:
                                DagsterEvent.engine_event(
                                    plan_context,
                                    "Executor received termination signal, not forwarding to steps"
                                    " because run will be resumed",
                                    EngineEventData(
                                        metadata={
                                            "steps_in_flight": MetadataValue.text(
                                                str(running_steps.keys())
                                            )
                                        },
                                    ),
                                )

                            return

                        if active_execution.has_in_flight_steps:
                            for dagster_event in self._pop_events(
                                plan_context.instance,
                                plan_context.run_id,
                                seen_storage_ids,
                            ):
                                yield dagster_event
                                # STEP_SKIPPED events are only emitted by ActiveExecution, which already handles
                                # and yields them.

                                if dagster_event.is_step_skipped:
                                    assert isinstance(dagster_event.step_key, str)
                                    active_execution.verify_complete(
                                        plan_context, dagster_event.step_key
                                    )
                                else:
                                    active_execution.handle_event(dagster_event)
                                    if (
                                        dagster_event.is_step_success
                                        or dagster_event.is_step_failure
                                        or dagster_event.is_resource_init_failure
                                        or dagster_event.is_step_up_for_retry
                                    ):
                                        assert isinstance(dagster_event.step_key, str)
                                        del running_steps[dagster_event.step_key]

                                        if not dagster_event.is_step_up_for_retry:
                                            active_execution.verify_complete(
                                                plan_context, dagster_event.step_key
                                            )

                        # process skips from failures or uncovered inputs
                        list(active_execution.plan_events_iterator(plan_context))

                        curr_time = get_current_datetime()
                        if (
                            curr_time - last_check_step_health_time
                        ).total_seconds() >= self._check_step_health_interval_seconds:
                            last_check_step_health_time = curr_time
                            for step in running_steps.values():
                                step_context = plan_context.for_step(step)

                                try:
                                    health_check_result = self._step_handler.check_step_health(
                                        self._get_step_handler_context(
                                            plan_context, [step], active_execution
                                        )
                                    )
                                    if not health_check_result.is_healthy:
                                        health_check_error = SerializableErrorInfo(
                                            message=f"Step {step.key} failed health check: {health_check_result.unhealthy_reason}",
                                            stack=[],
                                            cls_name=None,
                                        )

                                        self.get_failure_or_retry_event_after_crash(
                                            step_context,
                                            health_check_error,
                                            active_execution.get_known_state(),
                                        )

                                except Exception:
                                    serializable_error = serializable_error_info_from_exc_info(
                                        sys.exc_info()
                                    )
                                    # Log a step failure event if there was an error during the health
                                    # check
                                    DagsterEvent.step_failure_event(
                                        step_context=plan_context.for_step(step),
                                        step_failure_data=StepFailureData(
                                            error=serializable_error,
                                            user_failure_data=None,
                                        ),
                                    )

                        if self._max_concurrent is not None:
                            max_steps_to_run = self._max_concurrent - len(running_steps)
                            check.invariant(
                                max_steps_to_run >= 0, "More steps are active than max_concurrent"
                            )
                        else:
                            max_steps_to_run = None  # disables limit

                        # process events from concurrency blocked steps
                        list(active_execution.concurrency_event_iterator(plan_context))

                        for step in active_execution.get_steps_to_execute(max_steps_to_run):
                            running_steps[step.key] = step
                            list(
                                self._step_handler.launch_step(
                                    self._get_step_handler_context(
                                        plan_context, [step], active_execution
                                    )
                                )
                            )

                        time.sleep(self._sleep_seconds)
                except Exception:
                    if not active_execution.is_complete and running_steps:
                        serializable_error = serializable_error_info_from_exc_info(sys.exc_info())
                        DagsterEvent.engine_event(
                            plan_context,
                            "Unexpected exception while steps were still in-progress - terminating running steps:",
                            EngineEventData(
                                metadata={
                                    "steps_interrupted": MetadataValue.text(
                                        str(list(running_steps.keys()))
                                    )
                                },
                                error=serializable_error,
                            ),
                        )
                        for step in running_steps.values():
                            list(
                                self._step_handler.terminate_step(
                                    self._get_step_handler_context(
                                        plan_context, [step], active_execution
                                    )
                                )
                            )
                    raise
